import autoarray.plot as aplt


class Include1D(aplt.Include1D):
    def __init__(self, half_light_radius=None, einstein_radius=None):
        """
        Sets which `Visuals1D` are included on a figure plotting 1D data that is plotted using a `Plotter1D`.

        The `Include` object is used to extract the visuals of the plotted 1D data structures so they can be used in
        plot functions. Only visuals with a `True` entry in the `Include` object are extracted and plotted.

        If an entry is not input into the class (e.g. it retains its default entry of `None`) then the bool is
        loaded from the `config/visualize/include.ini` config file. This means the default visuals of a project
        can be specified in a config file.

        Parameters
        ----------
        half_light_radius
            If `True`, the `half_light_radius` of the plotted light profile is included on the figure.
        einstein_radius
            If `True`, the `einstein_radius` of the plotted mass profile is included on the figure.
        """
        super().__init__()

        self._half_light_radius = half_light_radius
        self._einstein_radius = einstein_radius

    @property
    def half_light_radius(self):
        return self.load(value=self._half_light_radius, name="half_light_radius")

    @property
    def einstein_radius(self):
        return self.load(value=self._einstein_radius, name="einstein_radius")



import numpy as np
from typing import Dict, List, Optional, Tuple, Type, Union

from autoconf import conf
import autoarray as aa

from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.profiles.basis import Basis
from autogalaxy.profiles.light.linear import LightProfileLinear
from autogalaxy.operate.image import OperateImageGalaxies
from autogalaxy.operate.deflections import OperateDeflections


class Galaxies(List, OperateImageGalaxies, OperateDeflections):
    def __init__(
        self,
        galaxies: List[Galaxy],
        run_time_dict: Optional[Dict] = None,
    ):
        """
        A collection of galaxies, used to perform operations on the galaxies as a group.

        It is common for a user to have multiple galaxies in a list, for which they may perform operations like
        creating the image of the light profiles or all galaxies or the potential of the mass profiles of all galaxies.

        Many of these calculations are straight forward, for example for the image of all galaxies we simply sum the
        images of each galaxy.

        However, there are more complex operations that can be performed on the galaxies as a group, for example
        computing the blured image of a group of galaxies where some galaxies have operated light profiles (meaning
        that PSF blurring is already applied to them and thus should be skipped) and another subset of galaxies have
        normal light profiles (meaning that PSF blurring should be applied to them).

        This calculation requires a careful set of operations to ensure that the PSF blurring is only applied to the
        subset of galaxies that do not have operated light profiles. This is an example of a calculation that is
        performed by the `Galaxies` class, simplifing the code required to perform this operation.

        Users may find that they often omit the `Galaxies` class and instead perform these operations in a more manual
        way. This is fine, but care must be taken to ensure that the operations are performed correctly.

        Parameters
        ----------
        galaxies
            The list of galaxies whose calculations are performed by this class.
        run_time_dict
            A dictionary of information on the run-times of function calls, including the total time and time spent on
            different calculations.
        """

        super().__init__(galaxies)
        self.run_time_dict = run_time_dict

    @property
    def redshift(self):
        return self[0].redshift

    def image_2d_list_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None
    ) -> List[aa.Array2D]:
        """
        Returns a list of the 2D images for each galaxy from a 2D grid of Cartesian (y,x) coordinates.

        The image of each galaxy is computed by summing the images of all light profiles in that galaxy. If a galaxy
        has no light profiles, a numpy array of zeros is returned.

        For example, if there are 3 galaxies and only the first two have light profiles, the returned list of images
        will be the image of the first two galaxies. The image of the third galaxies will be a numpy array of zeros.

        The images output by this function do not include instrument operations, such as PSF convolution (for imaging
        data) or a Fourier transform (for interferometer data).

        Inherited methods in the `autogalaxy.operate.image` package can apply these operations to the images.
        These functions may have the `operated_only` input passed to them, which is why this function includes
        the `operated_only` input.

        If the `operated_only` input is included, the function omits light profiles which are parents of
        the `LightProfileOperated` object, which signifies that the light profile represents emission that has
        already had the instrument operations (e.g. PSF convolution, a Fourier transform) applied to it and therefore
        that operation is not performed again.

        See the `autogalaxy.profiles.light` package for details of how images are computed from a light
        profile.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates where values of the image are evaluated.
        operated_only
            The returned list from this function contains all light profile images, and they are never operated on
            (e.g. via the imaging PSF). However, inherited methods in the `autogalaxy.operate.image` package can
            apply these operations to the images, which may have the `operated_only` input passed to them. This input
            therefore is used to pass the `operated_only` input to these methods.
        """
        return [
            galaxy.image_2d_from(grid=grid, operated_only=operated_only)
            for galaxy in self
        ]

    @aa.grid_dec.to_array
    def image_2d_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None
    ) -> aa.Array2D:
        """
        Returns the 2D image of all galaxies summed from a 2D grid of Cartesian (y,x) coordinates.

        This function first computes the image of each galaxy, via the function `image_2d_list_from`. The
        images are then summed to give the overall image of the galaxies.

        Refer to the function `image_2d_list_from` for a full description of the calculation and how the `operated_only`
        input is used.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates where values of the image are evaluated.
        operated_only
            The returned list from this function contains all light profile images, and they are never operated on
            (e.g. via the imaging PSF). However, inherited methods in the `autogalaxy.operate.image` package can
            apply these operations to the images, which may have the `operated_only` input passed to them. This input
            therefore is used to pass the `operated_only` input to these methods.
        """
        return sum(self.image_2d_list_from(grid=grid, operated_only=operated_only))

    def galaxy_image_2d_dict_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None
    ) -> {Galaxy: np.ndarray}:
        """
        Returns a dictionary associating every `Galaxy` object with its corresponding 2D image, using the instance
        of each galaxy as the dictionary keys.

        This object is used for adaptive-features, which use the image of each galaxy in a model-fit in order to
        adapt quantities like a pixelization or regularization scheme to the surface brightness of the galaxies being
        fitted.

        By inheriting from `OperateImageGalaxies` functions which apply operations of this dictionary are accessible,
        for example convolving every image with a PSF or applying a Fourier transform to create a galaxy-visibilities
        dictionary.

        Parameters
        ----------
        grid
            The 2D (y,x) coordinates of the (masked) grid, in its original geometric reference frame.

        Returns
        -------
        A dictionary associated every galaxy with its corresponding 2D image.
        """

        galaxy_image_2d_dict = dict()

        image_2d_list = self.image_2d_list_from(grid=grid, operated_only=operated_only)

        for galaxy_index, galaxy in enumerate(self):
            galaxy_image_2d_dict[galaxy] = image_2d_list[galaxy_index]

        return galaxy_image_2d_dict

    @aa.grid_dec.to_vector_yx
    def deflections_yx_2d_from(self, grid: aa.type.Grid2DLike, **kwargs) -> np.ndarray:
        """
        Returns the summed 2D deflections angles of all galaxies from a 2D grid of Cartesian (y,x) coordinates.

        The deflections of each galaxy is computed by summing the deflections of all mass profiles in that galaxy. If a
        galaxy has no mass profiles, a numpy array of zeros is returned.

        This calculation does not account for multi-plane ray-tracing effects, it is simply the sum of the deflections
        of all galaxies. The `Tracer` class in PyAutoLens is required for this.

        For example, if there are 3 galaxies and only the first two have mass profiles, the returned list of deflections
        will be the deflections of the first two galaxies. The deflections of the third galaxies will be a numpy
        array of zeros.

        See the `autogalaxy.profiles.mass` package for details of how deflections are computed from a mass profile.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates where values of the deflections are evaluated.
        """
        if self:
            return sum(map(lambda g: g.deflections_yx_2d_from(grid=grid), self))
        return np.zeros(shape=(grid.shape[0], 2))

    @aa.grid_dec.to_grid
    def traced_grid_2d_from(self, grid: aa.type.Grid2DLike) -> aa.type.Grid2DLike:
        """
        Trace this plane's grid_stacks to the next plane, using its deflection angles.
        """
        return grid - self.deflections_yx_2d_from(grid=grid)

    @aa.grid_dec.to_array
    def convergence_2d_from(self, grid: aa.type.Grid2DLike, **kwargs) -> np.ndarray:
        """
        Returns the summed 2D convergence of all galaxies from a 2D grid of Cartesian (y,x) coordinates.

        The convergence of each galaxy is computed by summing the convergence of all mass profiles in that galaxy. If a
        galaxy has no mass profiles, a numpy array of zeros is returned.

        This calculation does not account for multi-plane ray-tracing effects, it is simply the sum of the convergence
        of all galaxies. The `Tracer` class in PyAutoLens is required for this.

        For example, if there are 3 galaxies and only the first two have mass profiles, the returned list of convergence
        will be the convergence of the first two galaxies. The convergence of the third galaxies will be a numpy
        array of zeros.

        See the `autogalaxy.profiles.mass` package for details of how convergence are computed from a mass profile.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates where values of the convergence are evaluated.
        """
        if self:
            return sum(map(lambda g: g.convergence_2d_from(grid=grid), self))
        return np.zeros((grid.shape[0],))

    @aa.grid_dec.to_array
    def potential_2d_from(self, grid: aa.type.Grid2DLike, **kwargs) -> np.ndarray:
        """
        Returns the summed 2D potential of all galaxies from a 2D grid of Cartesian (y,x) coordinates.

        The potential of each galaxy is computed by summing the potential of all mass profiles in that galaxy. If a
        galaxy has no mass profiles, a numpy array of zeros is returned.

        This calculation does not account for multi-plane ray-tracing effects, it is simply the sum of the potential
        of all galaxies. The `Tracer` class in PyAutoLens is required for this.

        For example, if there are 3 galaxies and only the first two have mass profiles, the returned list of potential
        will be the potential of the first two galaxies. The potential of the third galaxies will be a numpy
        array of zeros.

        See the `autogalaxy.profiles.mass` package for details of how potential are computed from a mass profile.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates where values of the potential are evaluated.
        """
        if self:
            return sum(map(lambda g: g.potential_2d_from(grid=grid), self))
        return np.zeros((grid.shape[0],))

    def has(self, cls: Union[Type, Tuple[Type]]) -> bool:
        """
        Returns a bool specifying whether any of the galaxies has a certain class type.

        For example, for the input `cls=ag.LightProfile`, this function returns True if any galaxy has a
        light profile and false if no galaxy has a light profile.

        This function is used to check for mass profiles and specific types of profiles, like the linear light profile.

        Parameters
        ----------
        cls
            The class type of the galaxy which is checked for in the tracer.

        Returns
        -------
        True if any galaxy in the tracer has the input class type, else False.
        """
        return any(list(map(lambda galaxy: galaxy.has(cls=cls), self)))

    def cls_list_from(self, cls: Type) -> List:
        """
        Returns a list of objects in the galaxies which are an instance of the input `cls`.

        For example:

        - If the input is `cls=ag.LightProfile`, a list containing all light profiles of all galaxies is returned.

        Returns
        -------
            The list of objects in the galaxies that inherit from input `cls`.
        """
        cls_list = []

        for galaxy in self:
            if galaxy.has(cls=cls):
                for cls_galaxy in galaxy.cls_list_from(cls=cls):
                    cls_list.append(cls_galaxy)

        return cls_list

    def galaxies_with_cls_list_from(self, cls: Type) -> List[Galaxy]:
        return list(filter(lambda galaxy: galaxy.has(cls=cls), self))

    def galaxy_has_cls(self, cls: Type) -> bool:
        return any([galaxy.has(cls=cls) for galaxy in self])

    def extract_attribute(self, cls, attr_name):
        """
        Returns an attribute of a class in all galaxies as a `ValueIrregular` or `Grid2DIrregular` object.

        For example, if there is one galaxy with two light profiles and we want its axis-ratios, the following:

        `galaxies.extract_attribute(cls=LightProfile, name="axis_ratio")`

        would return:

        ArrayIrregular(values=[axis_ratio_0, axis_ratio_1])

        If a galaxy has three mass profiles and we want their centres, the following:

        `galaxies.extract_attribute(cls=MassProfile, name="centres")`

        would return:

        GridIrregular2D(grid=[(centre_y_0, centre_x_0), (centre_y_1, centre_x_1), (centre_y_2, centre_x_2)])

        This is used for visualization, for example plotting the centres of all mass profiles colored by their profile.
        """

        def extract(value, name):
            try:
                return getattr(value, name)
            except (AttributeError, IndexError):
                return None

        attributes = [
            extract(value, attr_name)
            for galaxy in self
            for value in galaxy.__dict__.values()
            if isinstance(value, cls)
        ]

        if attributes == []:
            return None
        elif isinstance(attributes[0], float):
            return aa.ArrayIrregular(values=attributes)
        elif isinstance(attributes[0], tuple):
            return aa.Grid2DIrregular(values=attributes)

    @property
    def perform_inversion(self) -> bool:
        """
        Returns a bool specifying whether this fit object performs an inversion.

        This is based on whether any of the galaxies have a `Pixelization` or `LightProfileLinear` object, in which
        case an inversion is performed.

        Returns
        -------
            A bool which is True if an inversion is performed.
        """
        if self.has(cls=(aa.Pixelization, LightProfileLinear)):
            return True
        elif self.has(cls=Basis):
            basis_list = self.cls_list_from(cls=Basis)
            for basis in basis_list:
                for light_profile in basis.profile_list:
                    if isinstance(light_profile, LightProfileLinear):
                        return True

        return False

    def plane_image_2d_from(
        self, grid: aa.type.Grid2DLike, zoom_to_brightest: bool = True
    ) -> aa.Array2D:
        return plane_image_from(
            galaxies=self,
            grid=grid.mask.derive_grid.all_false,
            zoom_to_brightest=zoom_to_brightest,
        )


def plane_image_from(
    galaxies: List[Galaxy],
    grid: aa.Grid2D,
    buffer: float = 1.0e-2,
    zoom_to_brightest: bool = True,
) -> aa.Array2D:
    """
    Returns the plane image of a list of galaxies, by summing their individual images.

    For lensing calculations performed by **PyAutoLens**, this function is used to return the unleensed image
    source-plane galaxies.

    By default, an adaptive grid is used to determine the grid that the images of the galaxies are computed on.
    This grid adapts its dimensions to capture the brightest regions of the image, ensuring that visualization of
    the plane-image is focused entirely on where the galaxies are brightest.

    This adaptive grid is based on determining the size of the grid that contains all pixels with an
    input % (typically 99%) of the total flux of the brightest pixel in the image.

    The adaptive grid can be disabled such that the input grid is used to compute the image of the galaxies.

    Parameters
    ----------
    galaxies
        The list of galaxies whose images are summed to compute the plane image.
    grid
        The grid of (y,x) coordinates for which the image of the galaxies is computed on, or from which the adaptive
        grid is derived.
    buffer
        The buffer around the adaptive grid that is used to ensure the image of the galaxies is not cut off.
    zoom_to_brightest
        If True, an adaptive grid is used to compute the image of the galaxies which zooms in on the brightest
        regions of the image. If False, the input grid is used.

    Returns
    -------
    The plane image of the galaxies, which is the sum of their individual images.
    """

    shape = grid.shape_native

    if zoom_to_brightest:
        try:
            image = sum(map(lambda g: g.image_2d_from(grid=grid), galaxies))
            image = image.native

            zoom_percent = conf.instance["visualize"]["general"]["zoom"][
                "plane_percent"
            ]

            fractional_value = np.max(image) * zoom_percent

            fractional_bool = image > fractional_value

            true_indices = np.argwhere(fractional_bool)

            y_max_pix = np.min(true_indices[:, 0])
            y_min_pix = np.max(true_indices[:, 0])
            x_min_pix = np.min(true_indices[:, 1])
            x_max_pix = np.max(true_indices[:, 1])

            grid = grid.native

            extent = (
                grid[0, x_min_pix][1] - buffer,
                grid[0, x_max_pix][1] + buffer,
                grid[y_min_pix, 0][0] - buffer,
                grid[y_max_pix, 0][0] + buffer,
            )

            extent = aa.util.geometry.extent_symmetric_from(extent=extent)

            pixel_scales = (
                float((extent[3] - extent[2]) / shape[0]),
                float((extent[1] - extent[0]) / shape[1]),
            )
            origin = ((extent[3] + extent[2]) / 2.0, (extent[1] + extent[0]) / 2.0)

            grid = aa.Grid2D.uniform(
                shape_native=grid.shape_native,
                pixel_scales=pixel_scales,
                origin=origin,
            )
        except ValueError:
            pass

    image = sum(map(lambda g: g.image_2d_from(grid=grid), galaxies))

    return aa.Array2D.no_mask(
        values=image.native, pixel_scales=grid.pixel_scales, origin=grid.origin
    )



import numpy as np
from scipy.integrate import quad
from typing import List, Tuple

import autoarray as aa

from autogalaxy.profiles.mass.stellar.sersic import AbstractSersic

from autogalaxy.profiles.mass.stellar.sersic import cse_settings_from


class SersicGradient(AbstractSersic):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        sersic_index
            Controls the concentration of the profile (lower -> less concentrated, higher -> more concentrated).
        mass_to_light_ratio
            The mass-to-light ratio of the light profile.
        mass_to_light_gradient
            The mass-to-light radial gradient.
        """
        super().__init__(
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )
        self.mass_to_light_gradient = mass_to_light_gradient

    @aa.grid_dec.to_vector_yx
    @aa.grid_dec.transform
    @aa.grid_dec.relocate_to_radial_minimum
    def deflections_2d_via_integral_from(self, grid: aa.type.Grid2DLike, **kwargs):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        def calculate_deflection_component(npow, index):
            sersic_constant = self.sersic_constant

            deflection_grid = self.axis_ratio * grid[:, index]

            for i in range(grid.shape[0]):
                deflection_grid[i] *= (
                    self.intensity
                    * self.mass_to_light_ratio
                    * quad(
                        self.deflection_func,
                        a=0.0,
                        b=1.0,
                        args=(
                            grid[i, 0],
                            grid[i, 1],
                            npow,
                            self.axis_ratio,
                            self.sersic_index,
                            self.effective_radius,
                            self.mass_to_light_gradient,
                            sersic_constant,
                        ),
                    )[0]
                )
            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotated_grid_from_reference_frame_from(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )

    @staticmethod
    def deflection_func(
        u,
        y,
        x,
        npow,
        axis_ratio,
        sersic_index,
        effective_radius,
        mass_to_light_gradient,
        sersic_constant,
    ):
        _eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x**2) + (y**2 / (1 - (1 - axis_ratio**2) * u))))
        )

        return (
            (((axis_ratio * _eta_u) / effective_radius) ** -mass_to_light_gradient)
            * np.exp(
                -sersic_constant
                * (((_eta_u / effective_radius) ** (1.0 / sersic_index)) - 1)
            )
            / ((1 - (1 - axis_ratio**2) * u) ** (npow + 0.5))
        )

    @aa.over_sample
    @aa.grid_dec.to_array
    @aa.grid_dec.transform
    @aa.grid_dec.relocate_to_radial_minimum
    def convergence_2d_from(self, grid: aa.type.Grid2DLike, **kwargs):
        """Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(
            self.eccentric_radii_grid_from(grid=grid, **kwargs)
        )

    def convergence_func(self, grid_radius: float) -> float:
        return (
            self.mass_to_light_ratio
            * (
                ((self.axis_ratio * grid_radius) / self.effective_radius)
                ** -self.mass_to_light_gradient
            )
            * self.image_2d_via_radii_from(grid_radius)
        )

    def decompose_convergence_via_mge(self):
        radii_min = self.effective_radius / 100.0
        radii_max = self.effective_radius * 20.0

        def sersic_gradient_2D(r):
            return (
                self.mass_to_light_ratio
                * self.intensity
                * (
                    ((self.axis_ratio * r) / self.effective_radius)
                    ** -self.mass_to_light_gradient
                )
                * np.exp(
                    -self.sersic_constant
                    * (((r / self.effective_radius) ** (1.0 / self.sersic_index)) - 1.0)
                )
            )

        return self._decompose_convergence_via_mge(
            func=sersic_gradient_2D, radii_min=radii_min, radii_max=radii_max
        )

    def decompose_convergence_via_cse(
        self, grid_radii: np.ndarray
    ) -> Tuple[List, List]:
        """
        Decompose the convergence of the Sersic profile into singular isothermal elliptical (sie) profiles.

        This decomposition uses the standard 2d profile of a Sersic mass profile.

        Parameters
        ----------
        func
            The function representing the profile that is decomposed into CSEs.
        radii_min:
            The minimum radius to fit
        radii_max:
            The maximum radius to fit
        total_sies
            The number of SIEs used to approximate the input func.
        sample_points: int (should be larger than 'total_sies')
            The number of data points to fit

        Returns
        -------
        Tuple[List, List]
            A list of amplitudes and core radii of every singular isothernal ellipsoids (sie) the mass profile is decomposed
            into.
        """

        upper_dex, lower_dex, total_cses, sample_points = cse_settings_from(
            effective_radius=self.effective_radius,
            sersic_index=self.sersic_index,
            sersic_constant=self.sersic_constant,
            mass_to_light_gradient=self.mass_to_light_gradient,
        )

        scaled_effective_radius = self.effective_radius / np.sqrt(self.axis_ratio)
        radii_min = scaled_effective_radius / 10.0**lower_dex
        radii_max = scaled_effective_radius * 10.0**upper_dex

        def sersic_gradient_2D(r):
            return (
                self.mass_to_light_ratio
                * self.intensity
                * (
                    ((self.axis_ratio * r) / scaled_effective_radius)
                    ** -self.mass_to_light_gradient
                )
                * np.exp(
                    -self.sersic_constant
                    * (
                        ((r / scaled_effective_radius) ** (1.0 / self.sersic_index))
                        - 1.0
                    )
                )
            )

        return self._decompose_convergence_via_cse_from(
            func=sersic_gradient_2D,
            radii_min=radii_min,
            radii_max=radii_max,
            total_cses=total_cses,
            sample_points=sample_points,
        )


class SersicGradientSph(SersicGradient):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        sersic_index
            Controls the concentration of the profile (lower -> less concentrated, higher -> more concentrated).
        mass_to_light_ratio
            The mass-to-light ratio of the light profile.
        mass_to_light_gradient
            The mass-to-light radial gradient.
        """
        super().__init__(
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )



from __future__ import annotations
from typing import Dict, List, Optional, Type, Union

from autoconf import cached_property

import autoarray as aa

from autoarray.inversion.pixelization.mappers.factory import mapper_from
from autoarray.inversion.inversion.factory import inversion_from
from autogalaxy.analysis.adapt_images.adapt_images import AdaptImages
from autogalaxy.profiles.light.linear import (
    LightProfileLinearObjFuncList,
)
from autogalaxy.profiles.basis import Basis
from autogalaxy.profiles.light.linear import LightProfileLinear
from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.galaxy.galaxies import Galaxies
from autogalaxy.analysis.preloads import Preloads


class AbstractToInversion:
    def __init__(
        self,
        dataset: Optional[Union[aa.Imaging, aa.Interferometer, aa.DatasetInterface]],
        adapt_images: Optional[AdaptImages] = None,
        settings_inversion: aa.SettingsInversion = aa.SettingsInversion(),
        preloads=Preloads(),
        run_time_dict: Optional[Dict] = None,
    ):
        """
        Abstract class which interfaces a dataset and input modeling object (e.g. galaxies, a tracer) with the
        inversion module. to setup a linear algebra calculation.

        The galaxies may contain linear light profiles whose `intensity` values are solved for via linear algebra in
        order to best-fit the data. In this case, this class extracts the linear light profiles, of all galaxies,
        computes their images and passes them to the `inversion` module such that they become the  `mapping_matrix`
        used in the linear algebra calculation.

        The galaxies may also contain pixelizations, which use a mesh (e.g. a Voronoi mesh) and regularization scheme
        to reconstruct the galaxy's light. This class extracts all pixelizations and uses them to set up `Mapper`
        objects which pair the dataset and pixelization to again set up the appropriate `mapping_matrix` and
        other linear algebra matrices (e.g. the `regularization_matrix`).

        This class does not perform the inversion or compute any of the linear algebra matrices itself. Instead,
        it acts as an interface between the dataset and galaxies and the inversion module, extracting the
        necessary information from galaxies and passing it to the inversion module.

        The modeling object may also contain standard light profiles which have an input `intensity` which is not
        solved for via linear algebra. These profiles should have already been evaluated and subtracted from the
        dataset before the inversion is performed. This is how an inversion is set up in the fit
        modules (e.g. `FitImaging`).

        Parameters
        ----------
        dataset
            The dataset containing the data which the inversion is performed on.
        adapt_images
            Images which certain pixelizations use to adapt their properties to the dataset, for example congregating
            the pixelization's pixels to the brightest regions of the image.
        settings_inversion
            The settings of the inversion, which controls how the linear algebra calculation is performed.
        preloads
            Preloads of the inversion, which are used to speed up the linear algebra calculation.
        run_time_dict
            A dictionary of run-time values used to compute the inversion, for example the noise-map normalization.
        """
        if dataset is not None:
            if dataset.noise_covariance_matrix is not None:
                raise aa.exc.InversionException(
                    """
                    You cannot perform an inversion (e.g. use a linear light profile or pixelization) 
                    if the dataset has a `noise_covariance_matrix`.
                    
                    This is because the linear algebra implementation is only valid under the assumption 
                    of independent gaussian noise.
                    """
                )

        self.dataset = dataset

        self.adapt_images = adapt_images

        self.settings_inversion = settings_inversion

        self.preloads = preloads
        self.run_time_dict = run_time_dict

    @property
    def convolver(self) -> Optional[aa.Convolver]:
        """
        Returns the convolver of the imaging dataset, if the inversion is performed on an imaging dataset.

        The `GalaxiesToInversion` class acts as an interface between the dataset and inversion module for
        both imaging and interferometer datasets. Only imaging datasets have a convolver, thus this property
        ensures that for an interferometer dataset code which references a convolver does not raise an error.

        Returns
        -------
        The convolver of the imaging dataset, if it is an imaging dataset.
        """
        try:
            return self.dataset.convolver
        except AttributeError:
            return None

    @property
    def transformer(self) -> Optional[Union[aa.TransformerNUFFT, aa.TransformerDFT]]:
        """
        Returns the transformer of the interferometer dataset, if the inversion is performed on an interferometer
        dataset.

        The `GalaxiesToInversion` class acts as an interface between the dataset and inversion module for
        both imaging and interferometer datasets. Only interferometer datasets have a transformer, thus this property
        ensures that for an imaging dataset code which references a transformer does not raise an error.

        Returns
        -------
        The transformer of the interferometer dataset, if it is an interferometer dataset.
        """
        try:
            return self.dataset.transformer
        except AttributeError:
            return None

    @property
    def border_relocator(self) -> Optional[aa.BorderRelocator]:
        """
        Returns the border relocator, which relocates pixels from the border of the inversion to the edge of the
        inversion, which is used to prevent edge effects in the reconstruction.

        A full description of the border relocator is given in the `BorderRelocator` class in PyAutoArray.

        Border relocation is only used if the `use_border_relocator` attribute is True in the `SettingsInversion` object.
        """
        if self.settings_inversion.use_border_relocator:
            return self.dataset.grids.border_relocator

    def cls_light_profile_func_list_galaxy_dict_from(
        self, cls: Type
    ) -> Dict[LightProfileLinearObjFuncList, Galaxy]:
        raise NotImplementedError

    @cached_property
    def lp_linear_func_list_galaxy_dict(
        self,
    ) -> Dict[LightProfileLinearObjFuncList, Galaxy]:
        raise NotImplementedError

    @cached_property
    def mapper_galaxy_dict(self):
        raise NotImplementedError

    @cached_property
    def linear_obj_galaxy_dict(
        self,
    ) -> Dict[Union[LightProfileLinearObjFuncList, aa.AbstractMapper], Galaxy]:
        """
        Returns a dictionary associating every linear object (e.g. a linear light profile or mapper) with the galaxy
        it belongs to.

        This object is primarily used by fit objects (e.g. `FitImaging`, `FitInterferometer`) after the
        inversion has been performed, to extract the individual reconstruct images of each galaxy.

        For example, if three galaxies are fitted via an inversion, we need to be able to extract the individual
        linear profiles from the inversion and associate them with the correct galaxy. This dictionary provides
        this association.

        Returns
        -------
        The dictionary associating linear objects with galaxies.
        """
        lp_linear_func_galaxy_dict = self.lp_linear_func_list_galaxy_dict

        mapper_galaxy_dict = self.mapper_galaxy_dict

        return {**lp_linear_func_galaxy_dict, **mapper_galaxy_dict}

    @cached_property
    def linear_obj_list(self) -> List[aa.LinearObj]:
        """
        Returns a list of linear objects (e.g. linear light profiles or mappers) which are used to perform the
        inversion.

        This list is passed directly to the inversion module, and thus acts as a the final interface between the
        galaxies and the inversion module.

        Returns
        -------
        The list of linear objects used to perform the inversion.
        """
        return list(self.linear_obj_galaxy_dict.keys())


class GalaxiesToInversion(AbstractToInversion):
    def __init__(
        self,
        dataset: Optional[Union[aa.Imaging, aa.Interferometer, aa.DatasetInterface]],
        galaxies: List[Galaxy],
        adapt_images: Optional[AdaptImages] = None,
        settings_inversion: aa.SettingsInversion = aa.SettingsInversion(),
        preloads=aa.Preloads(),
        run_time_dict: Optional[Dict] = None,
    ):
        """
        Interfaces a dataset and input list of galaxies with the inversion module. to setup a
        linear algebra calculation.

        The galaxies may contain linear light profiles whose `intensity` values are solved for via linear algebra in
        order to best-fit the data. In this case, this class extracts the linear light profiles, of all galaxies,
        computes their images and passes them to the `inversion` module such that they become the  `mapping_matrix`
        used in the linear algebra calculation.

        The galaxies may also contain pixelizations, which use a mesh (e.g. a Voronoi mesh) and regularization scheme
        to reconstruct the galaxy's light. This class extracts all pixelizations and uses them to set up `Mapper`
        objects which pair the dataset and pixelization to again set up the appropriate `mapping_matrix` and
        other linear algebra matrices (e.g. the `regularization_matrix`).

        This class does not perform the inversion or compute any of the linear algebra matrices itself. Instead,
        it acts as an interface between the dataset and galaxies and the inversion module, extracting the
        necessary information from galaxies and passing it to the inversion module.

        The galaxies may also contain standard light profiles which have an input `intensity` which is not solved for
        via linear algebra. These profiles should have already been evaluated and subtracted from the dataset before
        the inversion is performed. This is how an inversion is set up in the fit modules (e.g. `FitImaging`).

        Parameters
        ----------
        dataset
            The dataset containing the data which the inversion is performed on.
        galaxies
            The list of galaxies which are fitted to the dataset via the inversion.
        adapt_images
            Images which certain pixelizations use to adapt their properties to the dataset, for example congregating
            the pixelization's pixels to the brightest regions of the image.
        settings_inversion
            The settings of the inversion, which controls how the linear algebra calculation is performed.
        preloads
            Preloads of the inversion, which are used to speed up the linear algebra calculation.
        run_time_dict
            A dictionary of run-time values used to compute the inversion, for example the noise-map normalization.
        """
        self.galaxies = Galaxies(galaxies)

        super().__init__(
            dataset=dataset,
            adapt_images=adapt_images,
            settings_inversion=settings_inversion,
            preloads=preloads,
            run_time_dict=run_time_dict,
        )

    @property
    def has_mapper(self):
        if self.galaxies.has(cls=aa.Pixelization):
            return True

    def cls_light_profile_func_list_galaxy_dict_from(
        self, cls: Type
    ) -> Dict[LightProfileLinearObjFuncList, Galaxy]:
        """
        Returns a dictionary associating each list of linear light profiles with the galaxy they belong to.

        This function iterates over all galaxies and their light profiles, extracting their linear light profiles and
        for each galaxy grouping them into a `LightProfileLinearObjFuncList` object, which is associated with the
        galaxy via the dictionary.

        This `LightProfileLinearObjFuncList` object contains the attributes (e.g. the data `grid`, `light_profiles`)
        and functionality (e.g. a `mapping_matrix` method) that are required to perform the inversion. It is
        in this method that the `image_2d_from` method of each light profile is used to compute the linear algebra
        matrices.

        Special behaviour is implemented for the case where a galaxy has a `Basis` object, which contains a list
        of light profiles grouped into one object (e.g. when performing a multi Gaussian expansion where 30+ linear
        light profile Gaussians are often used). In this case, the function extracts all linear light profiles from
        the `Basis` object and associates them with the galaxy.

        It is expected that only two inputs are passed into the `cls` parameter, either `LightProfileLinear` or `Basis`.
        However, a specific linear light profile class (e.g. `Sersic`) could be input if extracting a specific type of
        light profile is desired.

        There is a noteable reason why a different `LightProfileLinearObjFuncList` object is created for each
        galaxy. In the project PyAutoLens, an inversion is performed on a strong lens, consisting of a lens galaxy
        and source galaxy whose grids are different (the source-plane grid is deflected by the lens galaxy).

        In this case, because each `LightProfileLinearObjFuncList` object is associated with a different galaxy
        it is also associated with a different grid. Users of PyAutoLens should therefore be aware that in
        the `lens.to_inversion` module multiple `GalaxiesToInversion` objects are used to setup the inversion in this
        way.

        Parameters
        ----------
        cls
            The class of the light profiles which are extracted from the galaxies, which is
            typically `LightProfileLinear` or `Basis`.

        Returns
        -------
        A dictionary associating each list of linear light profiles with the galaxy they belong to.
        """
        if not self.galaxies.has(cls=cls):
            return {}

        lp_linear_func_galaxy_dict = {}

        for galaxy in self.galaxies:
            if galaxy.has(cls=cls):
                for light_profile in galaxy.cls_list_from(cls=cls):
                    if isinstance(light_profile, LightProfileLinear):
                        light_profile_list = [light_profile]
                    else:
                        light_profile_list = light_profile.light_profile_list
                        light_profile_list = [
                            light_profile
                            for light_profile in light_profile_list
                            if isinstance(light_profile, LightProfileLinear)
                        ]

                    if len(light_profile_list) > 0:
                        lp_linear_func = LightProfileLinearObjFuncList(
                            grid=self.dataset.grids.uniform,
                            blurring_grid=self.dataset.grids.blurring,
                            convolver=self.dataset.convolver,
                            light_profile_list=light_profile_list,
                            regularization=light_profile.regularization,
                        )

                        lp_linear_func_galaxy_dict[lp_linear_func] = galaxy

        return lp_linear_func_galaxy_dict

    @cached_property
    def lp_linear_func_list_galaxy_dict(
        self,
    ) -> Dict[LightProfileLinearObjFuncList, Galaxy]:
        """
        Returns a dictionary associating each list of linear light profiles with the galaxy they belong to.

        You should first refer to the docstring of the `cls_light_profile_func_list_galaxy_dict_from` method for a
        description of this method.

        In brief, this method iterates over all galaxies and their light profiles, extracting their linear light profiles
        and for each galaxy grouping them into a `LightProfileLinearObjFuncList` object, which is associated with the
        galaxy via the dictionary. It also extracts linear light profiles from `Basis` objects and makes this
        associated.

        The `LightProfileLinearObjFuncList` object contains the attributes (e.g. the data `grid`, `light_profiles`)
        and functionality (e.g. a `mapping_matrix` method) that are required to perform the inversion.

        This function first creates a dictionary of linear light profiles associated with each galaxy, and then
        does the same for all `Basis` objects. The two dictionaries are then combined and returned.

        In the project PyAutoLens, this function is overwritten in order to fully account how ray tracing
        changes the images of each galaxy light profile. Users of PyAutoLens should therefore be aware that in
        the `lens.to_inversion` module this function behaves differently.

        Returns
        -------
        A dictionary associating each list of linear light profiles and basis objects with the galaxy they belong to.
        """
        lp_linear_light_profile_func_list_galaxy_dict = (
            self.cls_light_profile_func_list_galaxy_dict_from(cls=LightProfileLinear)
        )

        lp_basis_func_list_galaxy_dict = (
            self.cls_light_profile_func_list_galaxy_dict_from(cls=Basis)
        )

        return {
            **lp_linear_light_profile_func_list_galaxy_dict,
            **lp_basis_func_list_galaxy_dict,
        }

    @cached_property
    def image_plane_mesh_grid_list(
        self,
    ) -> Optional[List[aa.Grid2DIrregular]]:
        """
        Returns a list of image-plane mesh-grids, which are image-plane grids defining the centres of the pixels of
        the pixelization's mesh (e.g. the centres of Voronoi pixels).

        The `image_mesh` attribute of the pixelization object defines whether the centre of each mesh pixel are
        determined in the image-plane. When this is the case, the pixelization therefore has an image-plane mesh-grid,
        which needs to be computed before the inversion is performed.

        This function iterates over all galaxies with pixelizations, determines which pixelizations have an
        `image_mesh` and for these pixelizations computes the image-plane mesh-grid.

        It returns a list of all image-plane mesh-grids, which in the functions `mapper_from` and `mapper_galaxy_dict`
        are grouped into a `Mapper` object with other information required to perform the inversion using the
        pixelization.

        The order of this list is not important, because the `linear_obj_galaxy_dict` function associates each
        mapper object (and therefore image-plane mesh-grid) with the galaxy it belongs to and is therefore used
        elsewhere in the code (e.g. the fit module) to match inversion results to galaxies.

        Certain image meshes adapt their pixels to the dataset, for example congregating the pixels to the brightest
        regions of the image. This requires that `adapt_images` are used when setting up the image-plane mesh-grid.
        This function uses the `adapt_images` attribute of the `GalaxiesToInversion` object pass these images and
        raise an error if they are not present.

        Returns
        -------
        A list of image-plane mesh-grids, one for each pixelization with an image mesh.
        """
        if not self.galaxies.galaxy_has_cls(cls=aa.Pixelization):
            return None

        image_plane_mesh_grid_list = []

        for galaxy in self.galaxies.galaxies_with_cls_list_from(cls=aa.Pixelization):
            pixelization = galaxy.cls_list_from(cls=aa.Pixelization)[0]

            if pixelization.image_mesh is not None:
                try:
                    adapt_data = self.adapt_images.galaxy_image_dict[galaxy]
                except (AttributeError, KeyError):
                    adapt_data = None

                    if pixelization.image_mesh.uses_adapt_images:
                        raise aa.exc.PixelizationException(
                            """
                            Attempted to perform fit using a pixelization which requires an 
                            image-mesh (E.g. KMeans, Hilbert).
                            
                            However, the adapt-images passed to the fit (E.g. FitImaging, FitInterferometer) 
                            is None. Without an adapt image, an image-mesh cannot be used.
                            """
                        )

                image_plane_mesh_grid = (
                    pixelization.image_mesh.image_plane_mesh_grid_from(
                        mask=self.dataset.mask,
                        adapt_data=adapt_data,
                        settings=self.settings_inversion,
                    )
                )

            else:
                image_plane_mesh_grid = None

            image_plane_mesh_grid_list.append(image_plane_mesh_grid)

        return image_plane_mesh_grid_list

    def mapper_from(
        self,
        mesh: aa.AbstractMesh,
        regularization: aa.AbstractRegularization,
        source_plane_mesh_grid: aa.Grid2DIrregular,
        source_plane_data_grid: aa.Grid2D,
        adapt_galaxy_image: aa.Array2D,
        image_plane_mesh_grid: Optional[aa.Grid2DIrregular] = None,
    ) -> aa.AbstractMapper:
        """
        Returns a `Mapper` object from the attributes required to create one, which are extracted and computed
        from the dataset and galaxies.

        The `Mapper` object is used by a pixelization to perform an inversion. It maps pixels from the dataset's
        `data` to the pixels of the `mesh` which reconstruct the data via the inversion. These mappings are used to
        construct the `mapping_matrix` and other linear algebra matrices used in the inversion.

        This function is quite complex due to requirements from the child project PyAutoLens. In PyAutoLens, the
        `Mapper` object has grids corresponding to both the image-plane (e.g. the pixels of the dataset) and the
        source-plane (e.g. after gravitational lensing). There are also separate grids for the data and mesh pixels.
        In total, this means there are four grids: `image_plane_data_grid`, `image_plane_mesh_grid`,
        `source_plane_data_grid` and `source_plane_mesh_grid`. Lensing calculations are performed before these
        grids are passed  to this function.

        In PyAutoGalaxy, there is no lensing and therefore the `image_plane_data_grid` and `source_plane_data_grid`
        are identical, as are the `image_plane_mesh_grid` and `source_plane_mesh_grid`. This function therefore has
        an overly complex API, as it is designed to support PyAutoLens's use-cases.

        Parameters
        ----------
        mesh
            The mesh of the pixelization, which defines the pixels used to reconstruct the data (e.g. `Voronoi`).
        regularization
            The regularization scheme used to regularize the mesh pixel's reconstructed fluxes.
        source_plane_mesh_grid
            The mesh-grid of the source-plane which reconstructs the data (e.g. the centre of the `Voronoi` cells
            after lensing). In PyAutoGalaxy, this is identical to the `image_plane_mesh_grid`.
        source_plane_data_grid
            The data-grid of the source-plane, which are the ray-traced coordinates of the image-plane pixels
            that align with the image data. In PyAutoGalaxy, this is identical to the `image_plane_data_grid`.
        adapt_galaxy_image
            Images which certain pixelizations use to adapt their properties to the dataset, for example congregating
            the pixelization's pixels to the brightest regions of the image.
        image_plane_mesh_grid
            The mesh-grid of the image-plane, which are the centres of the pixels of the dataset. This is only required
            if the pixelization has an `image_mesh` attribute.

        Returns
        -------
        A `Mapper` object which maps the dataset's data to the pixelization's mesh.
        """
        mapper_grids = mesh.mapper_grids_from(
            mask=self.dataset.mask,
            border_relocator=self.border_relocator,
            source_plane_data_grid=source_plane_data_grid,
            source_plane_mesh_grid=source_plane_mesh_grid,
            image_plane_mesh_grid=image_plane_mesh_grid,
            adapt_data=adapt_galaxy_image,
            preloads=self.preloads,
            run_time_dict=self.run_time_dict,
        )

        return mapper_from(
            mapper_grids=mapper_grids,
            over_sampler=self.dataset.grids.pixelization.over_sampling.over_sampler_from(
                mask=self.dataset.mask
            ),
            regularization=regularization,
            run_time_dict=self.run_time_dict,
        )

    @cached_property
    def mapper_galaxy_dict(self) -> Dict[aa.AbstractMapper, Galaxy]:
        """
        Returns a dictionary associating each `Mapper` object with the galaxy it belongs to.

        The docstring of the function `mapper_from` describes the `Mapper` object in detail, and is used
        in this function to create the `Mapper` objects which are associated with the galaxies.

        In brief, the `Mappers` are used by pixelizations to perform an inversion. They map pixels from the dataset's
        `data` to the pixels of the `mesh` which reconstruct the data via the inversion. These mappings are used to
        construct the `mapping_matrix` and other linear algebra matrices used in the inversion.

        This function essentially finds all galaxies with pixelizations, performs all necessary calculations to
        set up the `Mapper` objects (e.g. compute the `image_plane_mesh_grid`), and then associates each `Mapper`
        with the galaxy it belongs to.

        Returns
        -------
        A dictionary associating each `Mapper` object with the galaxy it belongs to.
        """
        if not self.galaxies.galaxy_has_cls(cls=aa.Pixelization):
            return {}

        mesh_grid_list = self.image_plane_mesh_grid_list

        mapper_galaxy_dict = {}

        pixelization_list = []

        galaxies_with_pixelization_list = self.galaxies.galaxies_with_cls_list_from(
            cls=aa.Pixelization
        )

        for pix in self.galaxies.cls_list_from(cls=aa.Pixelization):
            pixelization_list.append(pix)

        for mapper_index in range(len(mesh_grid_list)):
            galaxy = galaxies_with_pixelization_list[mapper_index]

            try:
                adapt_galaxy_image = self.adapt_images.galaxy_image_dict[galaxy]
            except (AttributeError, KeyError):
                adapt_galaxy_image = None

            mapper = self.mapper_from(
                mesh=pixelization_list[mapper_index].mesh,
                regularization=pixelization_list[mapper_index].regularization,
                source_plane_data_grid=self.dataset.grids.pixelization.over_sampler.over_sampled_grid,
                source_plane_mesh_grid=mesh_grid_list[mapper_index],
                adapt_galaxy_image=adapt_galaxy_image,
                image_plane_mesh_grid=mesh_grid_list[mapper_index],
            )

            mapper_galaxy_dict[mapper] = galaxy

        return mapper_galaxy_dict

    @property
    def inversion(self) -> aa.AbstractInversion:
        """
        Returns an inversion object from the dataset, galaxies and inversion settings.

        The inversion uses all linear light profiles and pixelizations in the galaxies to fit the data.

        It solves for the linear light profile intensities and pixelization mesh pixel values via linear algebra,
        finding the solution which best fits the data after regularization is applied.

        The `GalaxiesToInversion` object acts as an interface between the dataset and galaxies and the inversion module,
        with many of its functions required to set up the inputs to the inversion object, primarily
        the `linear_obj_list` and `linear_obj_galaxy_dict` properties.

        Returns
        -------
        The inversion object which fits the dataset using the galaxies.
        """
        inversion = inversion_from(
            dataset=self.dataset,
            linear_obj_list=self.linear_obj_list,
            settings=self.settings_inversion,
            preloads=self.preloads,
            run_time_dict=self.run_time_dict,
        )

        inversion.linear_obj_galaxy_dict = self.linear_obj_galaxy_dict

        return inversion



import copy, matplotlib
import numpy as np
import matplotlib.pyplot as plt

from astropy import units
from matplotlib.colors import LinearSegmentedColormap


def subplot_fit(fit_list):
    # NOTE:
    logscale = False

    # NOTE:
    figure, axes = plt.subplots(nrows=1, ncols=2, figsize=(16, 8))

    image = fit.data.native
    image_temp = image

    # NOTE:
    # image = copy.copy(sample.image)
    # if sample.mask is not None:
    #     image[sample.mask.astype(bool)] = np.nan
    #
    #     image_temp = copy.copy(sample.image)
    #     image_temp[~sample.mask.astype(bool)] = np.nan
    # else:
    #     image_temp = None

    # NOTE:
    def custom_colormap():
        # Define the colors
        # colors = ['white', 'black', 'red']
        # positions = [0.0, 0.5, 1.0]
        colors = ["grey", "white", "red", "darkred", "black"]
        positions = [0.0, 0.25, 0.5, 0.75, 1.0]

        # Create the colormap
        cmap = LinearSegmentedColormap.from_list(
            "custom_colormap", list(zip(positions, colors))
        )

        return cmap

    vmin = None
    vmax = None

    if vmin is None:
        vmin = 0.025
    if vmax is None:
        vmax = 5.0

    norm = matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax)

    im = axes[0].imshow(
        np.log10(image) if logscale else image,
        cmap="jet",
        aspect="auto",
        norm=norm,
    )
    if image_temp is not None:
        axes[0].imshow(
            np.log10(image_temp) if logscale else image_temp,
            cmap="jet",
            aspect="auto",
            alpha=0.5,
        )

    ellipse_list = [fit.ellipse for fit in fit_list]

    list_of_angles = []
    list_of_y_fit = []
    list_of_y_errors_fit = []
    y_means = []
    y_stds = []

    radii

    for i, (a, parameters) in enumerate(zip(array, list_of_parameters)):
        if i == 0:
            m = sum(1 for value in parameters.values() if value is not None)

        y_fit, y_errors_fit, (x, y), angles = sample.extract(
            a=a, parameters=parameters, condition=extract_condition
        )
        list_of_angles.append(angles)
        list_of_y_fit.append(y_fit)
        list_of_y_errors_fit.append(y_errors_fit)

        if y_errors_fit is None:
            # y_errors_fit = 0.05 * y_fit
            raise NotImplementedError()

        # NOTE:
        y_mean = np.nanmean(y_fit)
        y_means.append(y_mean)

        # NOTE:
        y_std = np.nanstd(y_fit)
        y_stds.append(y_std)

        # NOTE:
        axes[0].plot(x, y, marker="o", markersize=2.5, color="w")

        # NOTE:
        axes[1].errorbar(
            angles * units.rad.to(units.deg),
            y_fit,
            yerr=y_errors_fit,
            linestyle="None",
            marker="o",
            markersize=2.5,
            color="black",
        )
        axes[1].axhline(y_mean, linestyle=":", color="black")

    levels = np.sort(np.log10(y_means)) if logscale else np.sort(y_means)
    axes[0].contour(
        np.log10(image) if logscale else image,
        # levels=y_means[::-1],
        levels=levels,
        colors="black",
    )
    colors = [im.cmap(im.norm(level)) for level in levels][::-1]

    for i, (angles, y_fit, y_errors_fit) in enumerate(
        zip(list_of_angles, list_of_y_fit, list_of_y_errors_fit)
    ):
        axes[1].errorbar(
            angles * units.rad.to(units.deg),
            y_fit,
            yerr=y_errors_fit,
            linestyle="None",
            marker="o",
            markersize=2.5,
            color=colors[i],
        )
    # axes[0].plot(
    #     [247],
    #     [250],
    #     marker="o"
    # )
    xticks = np.linspace(0, image.shape[1], 11)
    yticks = np.linspace(0, image.shape[0], 11)
    axes[0].set_xticks(xticks)
    axes[0].set_yticks(xticks)
    axes[1].set_xticks([0, 90, 180, 270, 360])
    axes[1].set_xlabel(r"$\phi$ (deg)", fontsize=15)
    axes[1].set_ylabel(r"$\rm I(\phi)$ [E/s]", fontsize=15)
    axes[1].tick_params(axis="y", labelsize=12.5)
    axes[1].yaxis.tick_right()
    axes[1].yaxis.set_label_position("right")
    for i, ax in enumerate(axes):
        ax.minorticks_on()
        ax.tick_params(
            axis="both",
            which="major",
            length=6,
            right=True,
            top=True,
            direction="in",
            colors="w" if i == 0 else "black",
        )
        ax.tick_params(
            axis="both",
            which="minor",
            length=3,
            right=True,
            top=True,
            direction="in",
            colors="w" if i == 0 else "black",
        )

    axes[1].set_yscale("log")

    # text = axes[0].text(
    #     0.05,
    #     0.95,
    #     "model 1",
    #     horizontalalignment='left',
    #     verticalalignment='center',
    #     transform=axes[0].transAxes,
    #     fontsize=25,
    #     weight="bold",
    #     color="w"
    # )
    # text.set_path_effects([patheffects.withStroke(linewidth=3, foreground='black')])

    # text = axes[0].text(
    #     0.7,
    #     0.95,
    #     "NGC 2274",
    #     horizontalalignment='left',
    #     verticalalignment='center',
    #     transform=axes[0].transAxes,
    #     fontsize=25,
    #     weight="bold",
    #     color="w"
    # )
    # text.set_path_effects([patheffects.withStroke(linewidth=3, foreground='black')])

    if a_max is None:
        axes[0].set_xlim(0, image.shape[1])
        axes[0].set_ylim(0, image.shape[0])
    else:
        x0 = list_of_parameters[0]["x0"]
        y0 = list_of_parameters[0]["y0"]
        axes[0].set_xlim(x0 - a_max, x0 + a_max)
        axes[0].set_ylim(y0 - a_max, y0 + a_max)

    # NOTE:
    figure.subplots_adjust(left=0.05, right=0.95, bottom=0.075, top=0.95, wspace=0.0)

    y = []
    for i, (a, chi_squares_i) in enumerate(zip(array, chi_squares)):
        n = len(chi_squares_i) - 1

        # NOTE:
        y_i = np.nansum(chi_squares_i[:-1]) / (n - m)
        y.append(y_i)

    figure_stats, axes_stats = plt.subplots()
    axes_stats.plot(array, y, linestyle="None", marker="o", color="black")
    axes_stats.set_xscale("log")
    axes_stats.set_yscale("log")
    # directory = "./MASSIVE/metadata"
    # filename = "{}/xy_model_default.numpy".format(directory)
    # with open(filename, 'wb') as f:
    #     np.save(f, [x, y])
    # plt.show()
    # exit()

    # NOTE:
    # chi_squares_flattened = list(itertools.chain(*chi_squares))
    # plt.hist(chi_squares_flattened, bins=100, alpha=0.75)

    # plt.show();exit()



import logging
from typing import Callable, Union

import autoarray as aa
import autofit as af
from autoconf import conf
from autofit.exc import PriorLimitException
from autogalaxy.analysis.preloads import Preloads

logger = logging.getLogger(__name__)

logger.setLevel(level="INFO")


class FitMaker:
    def __init__(self, model: af.Collection, fit_from: Callable):
        """
        Makes fits using an input PyAutoFit `model`, where the parameters of the model are drawn from its prior. This
        uses an input `fit_from`, which given an `instance` of the model creates the fit object.

        This is used for implicit preloading in the `Analysis` classes, whereby the created fits are compared against
        one another to determine whether certain components of the analysis can be preloaded.

        This includes functionality for creating the fit via the model in different ways, so that if certain
        models are ill-defined another is used instead.

        Parameters
        ----------
        model
            A **PyAutoFit** model object which via its parameters and their priors can created instances of the model.
        fit_from
            A function which given the instance of the model creates a `Fit` object.
        """

        self.model = model
        self.fit_from = fit_from

    @property
    def preloads_cls(self):
        return Preloads

    def fit_via_model_from(
        self, unit_value: float
    ) -> Union[aa.FitImaging, aa.FitInterferometer]:
        """
        Create a fit via the model.

        This first tries to compute the fit from the input `unit_value`, where the `unit_value` defines unit hyper
        cube values of each parameter's prior in the model, used to map each value to physical values for the fit.

        If this model fit produces an `Exception` because the parameter combination does not fit the data accurately,
        a sequence of random fits are instead used into an exception is not returned. However, if the number
        of `preload_attempts` defined in the configuration files is exceeded a None is returned.

        Parameters
        ----------
        unit_value
            The unit hyper cube values of each parameter's prior in the model, used to map each value to physical
            values for the fit.

        Returns
        -------
        fit
            A fit object where an instance of the model has been fitted to the data.
        """
        try:
            try:
                return self.fit_unit_instance_from(unit_value=unit_value)
            except IndexError as e:
                raise Exception from e
        except (Exception, PriorLimitException):
            return self.fit_random_instance_from()

    def fit_unit_instance_from(
        self, unit_value: float
    ) -> Union[aa.FitImaging, aa.FitInterferometer]:
        """
        Create a fit via the model using an input `unit_value`, where the `unit_value` defines unit hyper
        cube values of each parameter's prior in the model, used to map each value to physical values for the fit.

        Parameters
        ----------
        unit_value
            The unit hyper cube values of each parameter's prior in the model, used to map each value to physical
            values for the fit.

        Returns
        -------
        fit
            A fit object where an instance of the model has been fitted to the data.
        """
        instance = self.model.instance_from_unit_vector(
            unit_vector=[unit_value] * self.model.prior_count, ignore_prior_limits=True
        )

        fit = self.fit_from(
            instance=instance,
        )
        fit.figure_of_merit
        return fit

    def fit_random_instance_from(self) -> Union[aa.FitImaging, aa.FitInterferometer]:
        """
        Create a fit via the model by guessing a sequence of random fits until an exception is not returned. If
        the number of `preload_attempts` defined in the configuration files is exceeded a None is returned.

        Returns
        -------
        fit
            A fit object where an instance of the model has been fitted to the data.
        """

        preload_attempts = conf.instance["general"]["analysis"]["preload_attempts"]

        for i in range(preload_attempts):
            try:
                instance = self.model.random_instance(ignore_prior_limits=True)

                fit = self.fit_from(
                    instance=instance,
                )

                fit.figure_of_merit

                return fit

            except Exception as e:
                continue



import numpy as np
from functools import wraps
from typing import Optional, Union

import autoarray as aa


def check_operated_only(func):
    """
    Checks if a light profile is a `LightProfileOperated` class and therefore already has had operations like a
    PSF convolution performed.

    This is compared to the `only_operated` input to determine if the image of that light profile is returned, or
    an array of zeros.

    Parameters
    ----------
    func
        A function which checks the light profile class and determines what image is returned.

    Returns
    -------
        A function that returns a 2D image.
    """

    @wraps(func)
    def wrapper(
        obj,
        grid: aa.type.Grid1D2DLike,
        operated_only: Optional[bool] = None,
        *args,
        **kwargs
    ) -> Union[aa.Array2D, np.ndarray]:
        """
        This decorator checks if a light profile is a `LightProfileOperated` class and therefore already has had operations like a
        PSF convolution performed.

        This is compared to the `only_operated` input to determine if the image of that light profile is returned, or
        an array of zeros.

        Parameters
        ----------
        obj
            A light profile with an `image_2d_from` function whose class is inspected to determine if the image is
            operated on.
        grid
            A grid_like object of (y,x) coordinates on which the function values are evaluated.
        operated_only
            By default this is None and the image is returned irrespecive of light profile class (E.g. it does not matter
            if it is already operated or not). If this input is included as a bool, the light profile image is only
            returned if they are or are not already operated.

        Returns
        -------
            The 2D image, which is customized depending on whether it has been operated on.
        """

        from autogalaxy.profiles.light.operated import (
            LightProfileOperated,
        )

        if operated_only is None:
            return func(obj, grid, operated_only, *args, **kwargs)
        elif operated_only:
            if isinstance(obj, LightProfileOperated):
                return func(obj, grid, operated_only, *args, **kwargs)
            return np.zeros((grid.shape[0],))
        if not isinstance(obj, LightProfileOperated):
            return func(obj, grid, operated_only, *args, **kwargs)
        return np.zeros((grid.shape[0],))

    return wrapper



import json
import logging
from typing import Dict, List, Optional, Tuple, Union
from os import path
import os
import time

from autoconf import conf
import autofit as af
import autoarray as aa

from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.galaxy.galaxies import Galaxies
from autogalaxy.cosmology.lensing import LensingCosmology
from autogalaxy.cosmology.wrap import Planck15

logger = logging.getLogger(__name__)

logger.setLevel(level="INFO")


class Analysis(af.Analysis):
    def __init__(self, cosmology: LensingCosmology = Planck15):
        """
        Fits a model to a dataset via a non-linear search.

        This abstract Analysis class for all model-fits which fit galaxies, but does not perform a model-fit by
        itself (and is therefore only inherited from).

        This class stores the Cosmology used for the analysis and adapt images used for certain model classes.

        Parameters
        ----------
        cosmology
            The Cosmology assumed for this analysis.
        """
        self.cosmology = cosmology

    def galaxies_via_instance_from(
        self, instance: af.ModelInstance, run_time_dict: Optional[Dict] = None
    ) -> List[Galaxy]:
        """
        Create a list of galaxies from a model instance, which is used to fit the dataset.

        The instance may only contain galaxies, in which case this function is redundant. However, if the clumns
        API is being used, the instance will contain both galaxies and clumps, and they should be added to create
        the single list of galaxies used to fit the dataset.

        Parameters
        ----------
        instance
            An instance of the model that is fitted to the data by this analysis (whose parameters may have been set
            via a non-linear search).

        Returns
        -------
        A list of galaxies that is used to then fit the dataset.
        """
        if hasattr(instance, "clumps"):
            return Galaxies(
                galaxies=instance.galaxies + instance.clumps,
                run_time_dict=run_time_dict,
            )

        return Galaxies(galaxies=instance.galaxies, run_time_dict=run_time_dict)

    def dataset_model_via_instance_from(
        self, instance: af.ModelInstance
    ) -> aa.DatasetModel:
        """
        Create a dataset model from a model instance, which is used to fit the dataset.

        Parameters
        ----------
        instance
            An instance of the model that is fitted to the data by this analysis (whose parameters may have been set
            via a non-linear search).

        Returns
        -------
        A dataset_model that is used to then fit the dataset.
        """
        if hasattr(instance, "dataset_model"):
            return instance.dataset_model

    def make_result(
        self,
        samples_summary: af.SamplesSummary,
        paths: af.AbstractPaths,
        samples: Optional[af.SamplesPDF] = None,
        search_internal: Optional[object] = None,
        analysis: Optional[af.Analysis] = None,
    ) -> af.Result:
        """
        After the non-linear search is complete create its `Result`, which includes:

        - The samples of the non-linear search (E.g. MCMC chains, nested sampling samples) which are used to compute
          the maximum likelihood model, posteriors and other properties.

        - The model used to fit the data, which uses the samples to create specific instances of the model (e.g.
          an instance of the maximum log likelihood model).

        - The non-linear search used to perform the model fit.

        The `ResultImaging` object contains a number of methods which use the above objects to create the max
        log likelihood galaxies `FitImaging`, adapt-galaxy images,etc.

        Parameters
        ----------
        samples
            A PyAutoFit object which contains the samples of the non-linear search, for example the chains of an MCMC
            run of samples of the nested sampler.
        search
            The non-linear search used to perform this model-fit.

        Returns
        -------
        ResultImaging
            The result of fitting the model to the imaging dataset, via a non-linear search.
        """
        return self.Result(
            samples_summary=samples_summary,
            paths=paths,
            samples=samples,
            search_internal=search_internal,
            analysis=self,
        )

    def profile_log_likelihood_function(
        self, instance: af.ModelInstance, paths: Optional[af.DirectoryPaths] = None
    ) -> Tuple[Dict, Dict]:
        """
        This function is optionally called throughout a model-fit to profile the log likelihood function.

        All function calls inside the `log_likelihood_function` that are decorated with the `profile_func` are timed
        with their times stored in a dictionary called the `run_time_dict`.

        An `info_dict` is also created which stores information on aspects of the model and dataset that dictate
        run times, so the profiled times can be interpreted with this context.

        The results of this profiling are then output to hard-disk in the `preloads` folder of the model-fit results,
        which they can be inspected to ensure run-times are as expected.

        Parameters
        ----------
        instance
            An instance of the model that is being fitted to the data by this analysis (whose parameters have been set
            via a non-linear search).
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization and the pickled objects used by the aggregator output by this function.

        Returns
        -------
        Two dictionaries, the profiling dictionary and info dictionary, which contain the profiling times of the
        `log_likelihood_function` and information on the model and dataset used to perform the profiling.
        """

        if isinstance(paths, af.DatabasePaths):
            return

        run_time_dict = {}
        info_dict = {}

        repeats = conf.instance["general"]["profiling"]["repeats"]
        info_dict["repeats"] = repeats

        # Ensure numba functions are compiled before profiling begins.

        fit = self.fit_from(instance=instance)
        fit.figure_of_merit

        start = time.time()

        for _ in range(repeats):
            try:
                fit = self.fit_from(instance=instance)
                fit.figure_of_merit
            except Exception:
                logger.info(
                    "Profiling failed. Returning without outputting information."
                )
                return

        fit_time = (time.time() - start) / repeats

        run_time_dict["fit_time"] = fit_time

        fit = self.fit_from(instance=instance, run_time_dict=run_time_dict)
        fit.figure_of_merit

        try:
            info_dict["image_pixels"] = self.dataset.grids.uniform.shape_slim
            info_dict[
                "sub_total_light_profiles"
            ] = self.dataset.grids.uniform.over_sampler.sub_total
        except AttributeError:
            pass

        if fit.model_obj.has(cls=aa.Pixelization):
            info_dict["use_w_tilde"] = fit.inversion.settings.use_w_tilde
            try:
                info_dict[
                    "sub_total_pixelization"
                ] = self.dataset.grids.pixelization.over_sampler.sub_total
            except AttributeError:
                pass
            info_dict[
                "use_positive_only_solver"
            ] = fit.inversion.settings.use_positive_only_solver
            info_dict[
                "force_edge_pixels_to_zeros"
            ] = fit.inversion.settings.force_edge_pixels_to_zeros
            info_dict["use_w_tilde_numpy"] = fit.inversion.settings.use_w_tilde_numpy
            info_dict["source_pixels"] = len(fit.inversion.reconstruction)

            if hasattr(fit.inversion, "w_tilde"):
                info_dict[
                    "w_tilde_curvature_preload_size"
                ] = fit.inversion.w_tilde.curvature_preload.shape[0]

        self.output_profiling_info(
            paths=paths, run_time_dict=run_time_dict, info_dict=info_dict
        )

        return run_time_dict, info_dict

    def output_profiling_info(
        self, paths: Optional[af.DirectoryPaths], run_time_dict: Dict, info_dict: Dict
    ):
        """
        Output the log likelihood function profiling information to hard-disk as a json file.

        This function is separate from the `profile_log_likelihood_function` function above such that it can be
        called by children `Analysis` classes that profile additional aspects of the model-fit and therefore add
        extra information to the `run_time_dict` and `info_dict`.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization and the pickled objects used by the aggregator output by this function.
        run_time_dict
            A dictionary containing the profiling times of the functions called by the `log_likelihood_function`.
        info_dict
            A dictionary containing information on the model and dataset used to perform the profiling, where these
            settings typically control the overall run-time.
        """

        if paths is None:
            return

        os.makedirs(paths.profile_path, exist_ok=True)

        with open(path.join(paths.profile_path, "run_time_dict.json"), "w+") as f:
            json.dump(run_time_dict, f, indent=4)

        with open(path.join(paths.profile_path, "info_dict.json"), "w+") as f:
            json.dump(info_dict, f, indent=4)



from __future__ import annotations
import numpy as np
from typing import TYPE_CHECKING, Dict, Optional, Tuple

from autoconf import conf
from autoconf import cached_property

import autoarray as aa

if TYPE_CHECKING:
    from autogalaxy.galaxy.galaxy import Galaxy


class AdaptImages:
    def __init__(
        self,
        galaxy_image_dict: Optional[Dict[Galaxy, aa.Array2D]] = None,
        galaxy_name_image_dict: Optional[Dict[Tuple[str, ...], aa.Array2D]] = None,
    ):
        """
        Contains the adapt-images which are used to make a pixelization's mesh and regularization adapt to the
        reconstructed galaxy's morphology.

        Pixelization image-mesh objects (e.g. `KMeans`, `Hilbert`) adapt the distribution of pixels to the observed
        image's brightness and therefore to the reconstructed source's morphology.

        Certain regularization schemes (e.g. `AdaptiveBrightness`) adapt their regularization coefficients to the
        reconstructed source's morphology.

        These adaptive schemes use "adapt-images", which are images of each galaxy (e.g. the lens and source of a
        strong lens) estimated via an earlier model-fit.

        The adapt-images are stored as the model-image of each galaxy in a model (e.g. the lens and source for a
        strong lens). They are stored as a dictionary mapping each instance of the galaxy to its model-image.

        For model-fitting, the galaxy instances are updated for every iteration of the non-linear search. This means
        an `AdaptImages` instance cannot be passed directly to an `Analysis` class, as the galaxy instances need to be
        updated for every iteration of the non-linear search.

        A dictionary mapping the path name of each galaxy (e.g. "galaxies.lens") to its model-image is therefore used
        which is called inside the `log_likelihood_function` o map the model-image of each galaxy to the galaxy
        instance of that iteration's specific model.

        Parameters
        ----------
        galaxy_image_dict
            A dictionary associating each galaxy instance to an image of only that galaxy (e.g. for a strong lens
            one entry will map an instance of the source galaxy entry to an image of the lensed source.
        galaxy_name_image_dict
            A dictionary associating each galaxy path name (e.g. "galaxies.source") to an image of only that
            galaxy (e.g. for a strong lens the `source` entry is an image of the lensed source, without the lens light).
        """

        self.galaxy_image_dict = galaxy_image_dict
        self.galaxy_name_image_dict = galaxy_name_image_dict

    @property
    def mask(self) -> aa.Mask2D:
        """
        The mask of the adapt images.
        """
        try:
            return list(self.galaxy_image_dict.values())[0].mask
        except AttributeError:
            return list(self.galaxy_name_image_dict.values())[0].mask

    @cached_property
    def model_image(self) -> aa.Array2D:
        """
        The model-image is the sum of all individual galaxy images in the image dictionary.

        This is computed by summing the model-image of each individual adapt galaxy contained in the dictionary.
        """
        adapt_model_image = aa.Array2D(
            values=np.zeros(self.mask.pixels_in_mask),
            mask=self.mask,
        )

        try:
            for path in self.galaxy_image_dict.keys():
                adapt_model_image += self.galaxy_image_dict[path]
        except AttributeError:
            for path in self.galaxy_name_image_dict.keys():
                adapt_model_image += self.galaxy_name_image_dict[path]

        return adapt_model_image

    @classmethod
    def from_result(cls, result, use_model_images: bool = False) -> "AdaptImages":
        """
        Returns the adapt-images from a non-linear search result.

        For model-fitting, the adapt-images are typically setup using the maximum log likelihood model of the
        previous model-fit. This means the model-fitting is used to cleanly deblend the light of the different
        galaxies in the image (e.g. separate the lens light from the source light).

        This method uses attributes of a result (e.g. dictionary mapping galaxy instances to their model-images)
        to create the adapt-images.

        This can use either:

        - The model image of each galaxy in the best-fit model.
        - The subtracted image of each galaxy in the best-fit model, where the subtracted image is the dataset
          minus the model images of all other galaxies.

        Certain models produce galaxy-images with negative flux values (e.g. a pixelization), which can cause
        numerical issues with the adaptive schemes. To prevent this, we set a minimum flux value for each
        galaxy-image, which is a fraction of the maximum flux value of that image defined via a config file.

        Parameters
        ----------
        result
            The result of a previous model-fit, which contains the model-image of each galaxy.
        use_model_images
            If True, the model images of the galaxies are used to create the adapt images. If False, the subtracted
            images of the galaxies are used.

        Returns
        -------
        The adapt-images, which are the model-image of each galaxy inferred via the previous model-fit.
        """
        adapt_minimum_percent = conf.instance["general"]["adapt"][
            "adapt_minimum_percent"
        ]

        galaxy_name_image_dict = {}

        for path, galaxy in result.path_galaxy_tuples:
            if use_model_images:
                galaxy_image = result.model_image_galaxy_dict[path]
            else:
                galaxy_image = result.subtracted_signal_to_noise_map_galaxy_dict[path]

            if not np.all(galaxy_image == 0):
                minimum_galaxy_value = adapt_minimum_percent * max(galaxy_image)
                galaxy_image[galaxy_image < minimum_galaxy_value] = minimum_galaxy_value

            galaxy_name_image_dict[path] = galaxy_image

        return AdaptImages(galaxy_name_image_dict=galaxy_name_image_dict)

    def updated_via_instance_from(self, instance, mask=None) -> "AdaptImages":
        """
        Returns adapt-images which have been updated to map galaxy instances instead of galaxy names.

        For model-fitting, the galaxy instances are updated for every iteration of the non-linear search. This means
        an `AdaptImages` instance cannot be passed directly to an `Analysis` class, as the galaxy instances need to be
        updated for every iteration of the non-linear search.

        A dictionary mapping the path name of each galaxy (e.g. "galaxies.lens") to its model-image is therefore used
        which is called inside the `log_likelihood_function` o map the model-image of each galaxy to the galaxy
        instance of that iteration's specific model.

        This function is also called when loading an `AdaptImages` instance from a PyAutoFit database, as the
        galaxy instances are also created on-fly from the database. Database images do not have a mask, so it is
        also applied to the adapt images on-the-fly during database loading.

        Parameters
        ----------
        instance
            The instance of the model-fit (e.g. in a non-linear search) which is used to update the adapt images.
        mask
            A mask which can be applied to the adapt images, which is used when setting up the adaptive images
            via the aggregator and autofit database tools.

        Returns
        -------

        """
        from autogalaxy.galaxy.galaxy import Galaxy

        galaxy_image_dict = {}

        for galaxy_name, galaxy in instance.path_instance_tuples_for_class(Galaxy):
            galaxy_name = str(galaxy_name)

            if galaxy_name in self.galaxy_name_image_dict:
                galaxy_image_dict[galaxy] = self.galaxy_name_image_dict[galaxy_name]

        if mask is not None:
            for key, image in galaxy_image_dict.items():
                galaxy_image_dict[key] = aa.Array2D(values=image, mask=mask)

        return AdaptImages(galaxy_image_dict=galaxy_image_dict)



from astropy import cosmology as cosmo

from autogalaxy.cosmology.lensing import LensingCosmology


class Planck15(cosmo.FlatLambdaCDM, LensingCosmology):
    def __init__(self):
        """
        A wrapper for the astropy `Planck15` cosmology class.

        The only role of this class is to instantiate the `Planck15` cosmology class from astropy, but to additionally
        inherit from `LensingCosmology`, which is a class that provides additional functionality for calculating lensing
        specific quantities in the cosmology.
        """
        Planck15 = cosmo.Planck15

        super().__init__(
            H0=Planck15.H0,
            Om0=Planck15.Om0,
            Tcmb0=Planck15.Tcmb0,
            Neff=Planck15.Neff,
            m_nu=Planck15.m_nu,
            Ob0=Planck15.Ob0,
            name=Planck15.name,
        )



from astropy import cosmology as cosmo

from autogalaxy.cosmology.lensing import LensingCosmology


class LambdaCDMWrap(cosmo.LambdaCDM, LensingCosmology):
    def __init__(
        self,
        H0: float = 67.66,
        Om0: float = 0.30966,
        Ode0: float = 0.69034,
        Tcmb0: float = 2.7255,
        Neff: float = 3.046,
        m_nu: float = 0.06,
        Ob0: float = 0.04897,
    ):
        """
        A wrapper for the astropy `LambdaCDM` cosmology class, which allows it to be used for modeling such
        that the cosmological parameters are free parameters which can be fitted for.

        The interface of this class is the same as the astropy `LambdaCDM` class, it simply overwrites the
        __init__ method and inherits from it in a way that ensures **PyAutoFit** can compose a model from it
        without issue.

        The class also inherits from `LensingCosmology`, which is a class that provides additional functionality
        for calculating lensing specific quantities in the cosmology.

        Parameters
        ----------
        H0
            The Hubble constant at z=0.
        Om0
            The total matter density at z=0.
        Ode0
            The dark energy density at z=0.
        Tcmb0
            The CMB temperature at z=0.
        Neff
            The effective number of neutrinos.
        m_nu
            The sum of the neutrino masses.
        Ob0
            The baryon density at z=0.
        """
        super().__init__(
            H0=H0,
            Om0=Om0,
            Ode0=Ode0,
            Tcmb0=Tcmb0,
            Neff=Neff,
            m_nu=m_nu,
            Ob0=Ob0,
            name="FlatLambdaCDM",
        )


class FlatLambdaCDMWrap(cosmo.FlatLambdaCDM, LensingCosmology):
    def __init__(
        self,
        H0: float = 67.66,
        Om0: float = 0.30966,
        Tcmb0: float = 2.7255,
        Neff: float = 3.046,
        m_nu: float = 0.06,
        Ob0: float = 0.04897,
    ):
        """
        A wrapper for the astropy `FlatLambdaCDM` cosmology class, which allows it to be used for modeling such
        that the cosmological parameters are free parameters which can be fitted for.

        The interface of this class is the same as the astropy `FlatLambdaCDM` class, it simply overwrites the
        __init__ method and inherits from it in a way that ensures **PyAutoFit** can compose a model from it
        without issue.

        The class also inherits from `LensingCosmology`, which is a class that provides additional functionality
        for calculating lensing specific quantities in the cosmology.

        Parameters
        ----------
        H0
            The Hubble constant at z=0.
        Om0
            The total matter density at z=0.
        Tcmb0
            The CMB temperature at z=0.
        Neff
            The effective number of neutrinos.
        m_nu
            The sum of the neutrino masses.
        Ob0
            The baryon density at z=0.
        """
        super().__init__(
            H0=H0,
            Om0=Om0,
            Tcmb0=Tcmb0,
            Neff=Neff,
            m_nu=m_nu,
            Ob0=Ob0,
            name="FlatLambdaCDM",
        )


class FlatwCDMWrap(cosmo.FlatwCDM, LensingCosmology):
    def __init__(
        self,
        H0: float = 67.66,
        Om0: float = 0.30966,
        w0: float = -1.0,
        Tcmb0: float = 2.7255,
        Neff: float = 3.046,
        m_nu: float = 0.06,
        Ob0: float = 0.04897,
    ):
        """
        A wrapper for the astropy `FlatwCDM` cosmology class, which allows it to be used for modeling such
        that the cosmological parameters are free parameters which can be fitted for.

        The interface of this class is the same as the astropy `FlatwCDM` class, it simply overwrites the
        __init__ method and inherits from it in a way that ensures **PyAutoFit** can compose a model from it
        without issue.

        The class also inherits from `LensingCosmology`, which is a class that provides additional functionality
        for calculating lensing specific quantities in the cosmology.

        Parameters
        ----------
        H0
            The Hubble constant at z=0.
        Om0
            The total matter density at z=0.
        w0
            The dark energy equation of state at z=0.
        Tcmb0
            The CMB temperature at z=0.
        Neff
            The effective number of neutrinos.
        m_nu
            The sum of the neutrino masses.
        Ob0
            The baryon density at z=0.
        """
        super().__init__(
            H0=H0,
            Om0=Om0,
            w0=w0,
            Tcmb0=Tcmb0,
            Neff=Neff,
            m_nu=m_nu,
            Ob0=Ob0,
            name="FlatwCDM",
        )



import numpy as np
from typing import Optional, Tuple

import autoarray as aa

from autogalaxy.profiles.light.abstract import LightProfile
from autogalaxy.profiles.light.decorators import (
    check_operated_only,
)


class Moffat(LightProfile):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        alpha: float = 0.5,
        beta: float = 2.0,
    ):
        """
        The elliptical Moffat light profile, which is commonly used to model the Point Spread Function of
        Astronomy observations.

        This form of the MOffat profile is a reparameterizaiton of the original formalism given by
        https://ui.adsabs.harvard.edu/abs/1969A%26A.....3..455M/abstract. The actual profile itself is identical.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        alpha
            Scales the overall size of the Moffat profile and for a PSF typically corresponds to the FWHM / 2.
        beta
            Scales the wings at the outskirts of the Moffat profile, where smaller values imply heavier wings and it
            tends to a Gaussian as beta goes to infinity.
        """

        super().__init__(centre=centre, ell_comps=ell_comps, intensity=intensity)
        self.alpha = alpha
        self.beta = beta

    def image_2d_via_radii_from(self, grid_radii: np.ndarray) -> np.ndarray:
        """
        Returns the 2D image of the Moffat light profile from a grid of coordinates which are the radial distance of
        each coordinate from the its `centre`.

        Note: sigma is divided by sqrt(q) here.

        Parameters
        ----------
        grid_radii
            The radial distances from the centre of the profile, for each coordinate on the grid.
        """
        return np.multiply(
            self._intensity,
            np.power(
                1
                + np.square(
                    np.divide(grid_radii, self.alpha / np.sqrt(self.axis_ratio))
                ),
                -self.beta,
            ),
        )

    @aa.over_sample
    @aa.grid_dec.to_array
    @check_operated_only
    @aa.grid_dec.transform
    @aa.grid_dec.relocate_to_radial_minimum
    def image_2d_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None, **kwargs
    ) -> np.ndarray:
        """
        Returns the Moffat light profile's 2D image from a 2D grid of Cartesian (y,x) coordinates.

        If the coordinates have not been transformed to the profile's geometry (e.g. translated to the
        profile `centre`), this is performed automatically.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        image
            The image of the Moffat evaluated at every (y,x) coordinate on the transformed grid.
        """

        return self.image_2d_via_radii_from(
            self.eccentric_radii_grid_from(grid=grid, **kwargs)
        )


class MoffatSph(Moffat):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        alpha: float = 0.5,
        beta: float = 2.0,
    ):
        """
        The spherical Moffat light profile:

        This form of the MOffat profile is a reparameterizaiton of the original formalism given by
        https://ui.adsabs.harvard.edu/abs/1969A%26A.....3..455M/abstract. The actual profile itself is identical.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        alpha
            Scales the overall size of the Moffat profile and for a PSF typically corresponds to the FWHM / 2.
        beta
            Scales the wings at the outskirts of the Moffat profile, where smaller values imply heavier wings and it
            tends to a Gaussian as beta goes to infinity.
        """

        super().__init__(centre=centre, intensity=intensity, alpha=alpha, beta=beta)



from typing import Tuple

from autogalaxy.profiles.light.snr.abstract import LightProfileSNR
from autogalaxy.profiles.light import standard as lp


class Exponential(lp.Exponential, LightProfileSNR):
    def __init__(
        self,
        signal_to_noise_ratio: float = 10.0,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        effective_radius: float = 0.6,
    ):
        """
        The elliptical exponential profile.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 1.0.

        Parameters
        ----------
        centre
            The (y,x) arc-second centre of the light profile.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        """
        super().__init__(
            centre=centre,
            ell_comps=ell_comps,
            intensity=0.0,
            effective_radius=effective_radius,
        )
        LightProfileSNR.__init__(self, signal_to_noise_ratio=signal_to_noise_ratio)


class ExponentialSph(lp.ExponentialSph, LightProfileSNR):
    def __init__(
        self,
        signal_to_noise_ratio: float = 10.0,
        centre: Tuple[float, float] = (0.0, 0.0),
        effective_radius: float = 0.6,
    ):
        """
        The spherical exponential profile.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 1.0.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        """
        super().__init__(centre=centre, effective_radius=effective_radius)
        LightProfileSNR.__init__(self, signal_to_noise_ratio=signal_to_noise_ratio)



import numpy as np
from typing import Dict, List, Optional, Union

import autoarray as aa

from autogalaxy.profiles.light.abstract import LightProfile
from autogalaxy.profiles.mass.abstract.abstract import MassProfile

from autogalaxy.profiles.light import linear as lp_linear


class Basis(LightProfile, MassProfile):
    def __init__(
        self,
        profile_list: List[Union[LightProfile, MassProfile]],
        regularization: Optional[aa.AbstractRegularization] = None,
    ):
        """
        A basis is a collection of multiple light or mass profiles that are used to represent the light or mass
        distribution of a galaxy.

        A basis is typically used to decompose a galaxy's light or mass distribution into many profiles, each of
        which fit a small part of the overall light or mass distribution.

        For example, a common basis uses of order 10-100 Gaussian light or mass profiles, where each Gaussian
        represents a small part of the overall light or mass distribution of a galaxy. By decomposing the light or mass
        distribution into many profiles, more detailed structures can be captured and fitted that ordinary profiles
        would struggle to capture.

        This contrasts most standard light profiles (e.g. a Sersic) or mass profiles (e.g. an Isothermal) which
        represent the entire light or mass distribution of a galaxy.

        Parameters
        ----------
        profile_list
            The light or mass profiles that make up the basis.
        regularization
            The regularization scheme applied to the basis, which is used to regularize the solution to the linear
            inversion that fits the basis to the data.
        """
        super().__init__(
            centre=profile_list[0].centre,
            ell_comps=profile_list[0].ell_comps,
        )

        self.profile_list = profile_list
        self.regularization = regularization

    @property
    def light_profile_list(self) -> List[LightProfile]:
        """
        Returns a list of all light profiles in the `Basis` object.

        This is used for computing light profile quantities of each individual light profile in the `Basis` object and
        then summing them to get the overall quantity (e.g. the image, surface brightness, etc.).

        Returns
        -------
            The list of light profiles in the `Basis` object.
        """
        return aa.util.misc.cls_list_from(values=self.profile_list, cls=LightProfile)

    @property
    def mass_profile_list(self) -> List[MassProfile]:
        """
        Returns a list of all mass profiles in the `Basis` object.

        This is used for computing mass profile quantities of each individual mass profile in the `Basis` object and
        then summing them to get the overall quantity (e.g. the convergence, potential, etc.).

        Returns
        -------
            The list of mass profiles in the `Basis` object.
        """
        return aa.util.misc.cls_list_from(values=self.profile_list, cls=MassProfile)

    def image_2d_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None, **kwargs
    ) -> aa.Array2D:
        """
        Returns the summed image of all light profiles in the basis from a 2D grid of Cartesian (y,x) coordinates.

        Normal steps in the calculation of an image, like shifting the input grid to the profile's centre, rotating
        it to its position angle, and checking if its already operated on are all handled internally by
        each profiles `image_2d_from` method when it is called.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.
        operated_only
            By default, the returned list contains all light profile images (irrespective of whether they have been
            operated on or not). If this input is included as a bool, only images which are or are not already
            operated are included in the list, with the images of other light profiles created as a numpy array of
            zeros.

        Returns
        -------
        The image of the light profiles in the basis summed together.
        """
        return sum(self.image_2d_list_from(grid=grid, operated_only=operated_only))

    def image_2d_list_from(
        self, grid: aa.type.Grid2DLike, operated_only: Optional[bool] = None
    ) -> List[aa.Array2D]:
        """
        Returns each image of each light profiles in the basis as a list, from a 2D grid of Cartesian (y,x) coordinates.

        Normal steps in the calculation of an image, like shifting the input grid to the profile's centre, rotating
        it to its position angle, and checking if its already operated on are all handled internally by
        each profiles `image_2d_from` method when it is called.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.
        operated_only
            By default, the returned list contains all light profile images (irrespective of whether they have been
            operated on or not). If this input is included as a bool, only images which are or are not already
            operated are included in the list, with the images of other light profiles created as a numpy array of
            zeros.

        Returns
        -------
        The image of the light profiles in the basis summed together.
        """
        return [
            light_profile.image_2d_from(grid=grid, operated_only=operated_only)
            if not isinstance(light_profile, lp_linear.LightProfileLinear)
            else np.zeros((grid.shape[0],))
            for light_profile in self.light_profile_list
        ]

    def convergence_2d_from(self, grid: aa.type.Grid2DLike, **kwargs) -> aa.Array2D:
        """
        Returns the summed convergence of all mass profiles in the basis from a 2D grid of Cartesian (y,x) coordinates.

        Normal steps in the calculation of a convergence, like shifting the input grid to the profile's centre and
        rotating it to its position angle are all handled internally by each profile's `convergence_2d_from` method
        when it is called.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        The convergence of the mass profiles in the basis summed together.
        """
        if len(self.mass_profile_list) > 0:
            return sum(
                [mass.convergence_2d_from(grid=grid) for mass in self.mass_profile_list]
            )
        return np.zeros((grid.shape[0],))

    def potential_2d_from(self, grid: aa.type.Grid2DLike, **kwargs) -> aa.Array2D:
        """
        Returns the summed potential of all mass profiles in the basis from a 2D grid of Cartesian (y,x) coordinates.

        Normal steps in the calculation of a potential, like shifting the input grid to the profile's centre and
        rotating it to its position angle are all handled internally by each profile's `potential_2d_from` method
        when it is called.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        The potential of the mass profiles in the basis summed together.
        """
        if len(self.mass_profile_list) > 0:
            return sum(
                [mass.potential_2d_from(grid=grid) for mass in self.profile_list]
            )
        return np.zeros((grid.shape[0],))

    def deflections_yx_2d_from(self, grid: aa.type.Grid2DLike, **kwargs) -> aa.Array2D:
        """
        Returns the summed deflections of all mass profiles in the basis from a 2D grid of Cartesian (y,x) coordinates.

        Normal steps in the calculation of a deflections, like shifting the input grid to the profile's centre and
        rotating it to its position angle are all handled internally by each profile's `deflections_2d_from` method
        when it is called.

        Parameters
        ----------
        grid
            The 2D (y, x) coordinates in the original reference frame of the grid.

        Returns
        -------
        The deflections of the mass profiles in the basis summed together.
        """
        if len(self.mass_profile_list) > 0:
            return sum(
                [mass.deflections_yx_2d_from(grid=grid) for mass in self.profile_list]
            )
        return np.zeros((grid.shape[0], 2))

    def lp_instance_from(self, linear_light_profile_intensity_dict: Dict):
        light_profile_list = []

        for light_profile in self.profile_list:
            if isinstance(light_profile, lp_linear.LightProfileLinear):
                light_profile = light_profile.lp_instance_from(
                    linear_light_profile_intensity_dict=linear_light_profile_intensity_dict
                )

            light_profile_list.append(light_profile)

        return Basis(
            profile_list=light_profile_list, regularization=self.regularization
        )



from astropy import units
import numpy as np
from typing import Tuple


def ell_comps_from(axis_ratio: float, angle: float) -> Tuple[float, float]:
    """
    Returns the elliptical components e1 and e2 of a light or mass profile from an input angle in degrees and axis
    ratio.

    The elliptical components of a light or mass profile are given by:

    elliptical_component_y = ell_comps[0] = (1-axis_ratio)/(1+axis_ratio) * sin(2 * angle)
    elliptical_component_x = ell_comps[1] = (1-axis_ratio)/(1+axis_ratio) * cos(2 * angle)

    Which are the values this function returns.

    Parameters
    ----------
    axis_ratio
        Ratio of light profiles ellipse's minor and major axes (b/a).
    angle
        Rotation angle of light profile counter-clockwise from positive x-axis.
    """
    angle *= np.pi / 180.0
    fac = (1 - axis_ratio) / (1 + axis_ratio)
    ellip_y = fac * np.sin(2 * angle)
    ellip_x = fac * np.cos(2 * angle)
    return (ellip_y, ellip_x)


def axis_ratio_and_angle_from(ell_comps: Tuple[float, float]) -> Tuple[float, float]:
    """
    Returns the axis-ratio and position angle in degrees (-45 < angle < 135.0) from input elliptical components e1
    and e2 of a light or mass profile.

    The elliptical components of a light or mass profile are given by:

    elliptical_component_y = ell_comps[0] = (1-axis_ratio)/(1+axis_ratio) * sin(2 * angle)
    elliptical_component_x = ell_comps[1] = (1-axis_ratio)/(1+axis_ratio) * cos(2 * angle)

    The axis-ratio and angle are therefore given by:

    axis_ratio = (1 - fac) / (1 + fac)
    angle = 0.5 * arctan(ell_comps[0] / ell_comps[1])

    where `fac = sqrt(ell_comps[1] ** 2 + ell_comps[0] ** 2).

    This function returns the axis-ratio and angle in degrees.

    An additional check is performed which requires the angle is between -45 and 135 degrees. This ensures that
    for certain values of `ell_comps` the angle does not jump from one boundary to another (e.g. without this check
    certain values of `ell_comps` return -1.0 degrees and others 179.0 degrees). This ensures that when error
    estimates are computed from samples of a lens model via marginalization, the calculation is not biased by the
    angle jumping between these two values.

    Parameters
    ----------
    ell_comps
        The elliptical components of the light or mass profile which are converted to an angle.
    """
    angle = np.arctan2(ell_comps[0], ell_comps[1]) / 2
    angle *= 180.0 / np.pi

    if abs(angle) > 45 and angle < 0:
        angle += 180

    fac = np.sqrt(ell_comps[1] ** 2 + ell_comps[0] ** 2)
    if fac > 0.999:
        fac = 0.999  # avoid unphysical solution
    # if fac > 1: print('unphysical e1,e2')
    axis_ratio = (1 - fac) / (1 + fac)
    return axis_ratio, angle


def axis_ratio_from(ell_comps: Tuple[float, float]):
    """
    Returns the axis-ratio from input elliptical components e1 and e2 of a light or mass profile.

    The elliptical components of a light or mass profile are given by:

    elliptical_component_y = ell_comps[0] = (1-axis_ratio)/(1+axis_ratio) * sin(2 * angle)
    elliptical_component_x = ell_comps[1] = (1-axis_ratio)/(1+axis_ratio) * cos(2 * angle)

    The axis-ratio is therefore given by:

    axis_ratio = (1 - fac) / (1 + fac)
    angle = 0.5 * arctan(ell_comps[0] / ell_comps[1])

    where `fac = sqrt(ell_comps[1] ** 2 + ell_comps[0] ** 2).

    This function returns the axis-ratio.

    Parameters
    ----------
    ell_comps
        The elliptical components of the light or mass profile which are converted to an angle.
    """
    axis_ratio, angle = axis_ratio_and_angle_from(ell_comps=ell_comps)
    return axis_ratio


def angle_from(ell_comps: Tuple[float, float]) -> float:
    """
    Returns the position angle in degrees (-45 < angle < 135.0) from input elliptical components e1 and e2
    of a light or mass profile.

    The elliptical components of a light or mass profile are given by:

    elliptical_component_y = ell_comps[0] = (1-q)/(1+q) * sin(2 * angle)
    elliptical_component_x = ell_comps[1] = (1-q)/(1+q) * cos(2 * angle)

    axis_ratio = (1 - fac) / (1 + fac)
    angle = 0.5 * arctan(ell_comps[0] / ell_comps[1])

    where `fac = sqrt(ell_comps[1] ** 2 + ell_comps[0] ** 2).

    This function returns the angle in degrees.

    An additional check is performed which requires the angle is between -45 and 135 degrees. This ensures that
    for certain values of `ell_comps` the angle does not jump from one boundary to another (e.g. without this check
    certain values of `ell_comps` return -1.0 degrees and others 179.0 degrees).

    This ensures that when error estimates are computed from samples of a lens model via marginalization, the
    calculation is not biased by the angle jumping between these two values.

    Parameters
    ----------
    ell_comps
        The elliptical components of the light or mass profile which are converted to an angle.
    """
    axis_ratio, angle = axis_ratio_and_angle_from(ell_comps=ell_comps)

    return angle


def shear_gamma_1_2_from(magnitude: float, angle: float) -> Tuple[float, float]:
    """
    Returns the shear gamma 1 and gamma 2 values an input shear magnitude and angle in degrees.

    The gamma 1 and gamma 2 components of a shear are given by:

    gamma_1 = magnitude * np.cos(2 * angle * np.pi / 180.0)
    gamma_2 = magnitude * np.sin(2 * angle * np.pi / 180.0)

    Which are the values this function returns.

    Converting from gamma 1 and gamma 2 to magnitude and angle is given by:

    magnitude = np.sqrt(gamma_1**2 + gamma_2**2)
    angle = np.arctan2(gamma_2, gamma_1) / 2 * 180.0 / np.pi

    Parameters
    ----------
    axis_ratio
        Ratio of light profiles ellipse's minor and major axes (b/a).
    angle
        Rotation angle of light profile counter-clockwise from positive x-axis.
    """
    gamma_1 = magnitude * np.cos(2 * angle * np.pi / 180.0)
    gamma_2 = magnitude * np.sin(2 * angle * np.pi / 180.0)
    return (gamma_1, gamma_2)


def shear_magnitude_and_angle_from(
    gamma_1: float, gamma_2: float
) -> Tuple[float, float]:
    """
    Returns the shear magnitude and angle in degrees from input shear gamma 1 and gamma 2 values.

    The gamma 1 and gamma 2 components of a shear are given by:

    gamma_1 = magnitude * np.cos(2 * angle * np.pi / 180.0)
    gamma_2 = magnitude * np.sin(2 * angle * np.pi / 180.0)

    Converting from gamma 1 and gamma 2 to magnitude and angle is given by:

    magnitude = np.sqrt(gamma_1**2 + gamma_2**2)
    angle = np.arctan2(gamma_2, gamma_1) / 2 * 180.0 / np.pi

    Which are the values this function returns.

    Additional checks are performed which requires the angle is between -45 and 135 degrees. This ensures that
    for certain values of `gamma_1` and `gamma_2` the angle does not jump from one boundary to another (e.g. without
    this check certain values of `gamma_1` and `gamma_2` return -1.0 degrees and others 179.0 degrees).

    This ensures that when error estimates are computed from samples of a lens model via marginalization, the
    calculation is not biased by the angle jumping between these two values.

    Parameters
    ----------
    gamma_1
        The gamma 1 component of the shear.
    gamma_2
        The gamma 2 component of the shear.
    """
    angle = np.arctan2(gamma_2, gamma_1) / 2 * 180.0 / np.pi
    magnitude = np.sqrt(gamma_1**2 + gamma_2**2)

    if angle < 0:
        angle += 180.0

    if abs(angle - 90) > 45 and angle > 90:
        angle -= 180

    return magnitude, angle


def shear_magnitude_from(gamma_1: float, gamma_2: float) -> float:
    """
    Returns the shear magnitude and angle in degrees from input shear gamma 1 and gamma 2 values.

    The gamma 1 and gamma 2 components of a shear are given by:

    gamma_1 = magnitude * np.cos(2 * angle * np.pi / 180.0)
    gamma_2 = magnitude * np.sin(2 * angle * np.pi / 180.0)

    Converting from gamma 1 and gamma 2 to magnitude and angle is given by:

    magnitude = np.sqrt(gamma_1**2 + gamma_2**2)
    angle = np.arctan2(gamma_2, gamma_1) / 2 * 180.0 / np.pi

    The magnitude value is what this function returns.

    Parameters
    ----------
    gamma_1
        The gamma 1 component of the shear.
    gamma_2
        The gamma 2 component of the shear.
    """
    magnitude, angle = shear_magnitude_and_angle_from(gamma_1=gamma_1, gamma_2=gamma_2)
    return magnitude


def shear_angle_from(gamma_1: float, gamma_2: float) -> float:
    """
    Returns the shear magnitude and angle in degrees from input shear gamma 1 and gamma 2 values.

    The gamma 1 and gamma 2 components of a shear are given by:

    gamma_1 = magnitude * np.cos(2 * angle * np.pi / 180.0)
    gamma_2 = magnitude * np.sin(2 * angle * np.pi / 180.0)

    Converting from gamma 1 and gamma 2 to magnitude and angle is given by:

    magnitude = np.sqrt(gamma_1**2 + gamma_2**2)
    angle = np.arctan2(gamma_2, gamma_1) / 2 * 180.0 / np.pi

    The angle value is what this function returns.

    Parameters
    ----------
    gamma_1
        The gamma 1 component of the shear.
    gamma_2
        The gamma 2 component of the shear.
    """
    magnitude, angle = shear_magnitude_and_angle_from(gamma_1=gamma_1, gamma_2=gamma_2)
    return angle


def multipole_k_m_and_phi_m_from(
    multipole_comps: Tuple[float, float], m: int
) -> Tuple[float, float]:
    """
    Returns the multipole normalization value `k_m` and angle `phi` from the multipole component parameters.

    The normalization and angle are given by:

    .. math::
        \phi^{\rm mass}_m = \arctan{\frac{\epsilon_{\rm 2}^{\rm mp}}{\epsilon_{\rm 2}^{\rm mp}}}, \, \,
        k^{\rm mass}_m = \sqrt{{\epsilon_{\rm 1}^{\rm mp}}^2 + {\epsilon_{\rm 2}^{\rm mp}}^2} \, .

    The conversion depends on the multipole order `m`, to ensure that all possible rotationally symmetric
    multiple mass profiles are available in the conversion for multiple components spanning -inf to inf.

    Additional checks are performed which requires the angle `phi_m` is between -45 and 135 degrees. This ensures that
    for certain multipole component values the angle does not jump from one boundary to another (e.g. without
    this check certain values of `gamma_1` and `gamma_2` return -1.0 degrees and others 179.0 degrees).

    This ensures that when error estimates are computed from samples of a lens model via marginalization, the
    calculation is not biased by the angle jumping between these two values.

    Parameters
    ----------
    multipole_comps
        The first and second components of the multipole.

    Returns
    -------
    The normalization and angle parameters of the multipole.
    """
    phi_m = (
        np.arctan2(multipole_comps[0], multipole_comps[1]) * 180.0 / np.pi / float(m)
    )
    k_m = np.sqrt(multipole_comps[1] ** 2 + multipole_comps[0] ** 2)

    if phi_m < -90.0 / m:
        phi_m += 360.0 / m

    return k_m, phi_m


def multipole_comps_from(k_m: float, phi_m: float, m: int) -> Tuple[float, float]:
    """
    Returns the multipole component parameters from their normalization value `k_m` and angle `phi`.

    .. math::
        \phi^{\rm mass}_m = \arctan{\frac{\epsilon_{\rm 2}^{\rm mp}}{\epsilon_{\rm 2}^{\rm mp}}}, \, \,
        k^{\rm mass}_m = \sqrt{{\epsilon_{\rm 1}^{\rm mp}}^2 + {\epsilon_{\rm 2}^{\rm mp}}^2} \, .

    The conversion depends on the multipole order `m`, to ensure that all possible rotationally symmetric
    multiple mass profiles are available in the conversion for multiple components spanning -inf to inf.

    Parameters
    ----------
    k_m
        The magnitude of the multipole.
    phi_m
        The angle of the multipole.

    Returns
    -------
    The multipole component parameters.
    """
    multipole_comp_0 = k_m * np.sin(phi_m * float(m) * units.deg.to(units.rad))
    multipole_comp_1 = k_m * np.cos(phi_m * float(m) * units.deg.to(units.rad))

    return (multipole_comp_0, multipole_comp_1)



import autoarray as aa

from autoarray.fit.plot.fit_imaging_plotters import FitImagingPlotterMeta

from autogalaxy.quantity.fit_quantity import FitQuantity

from autogalaxy.plot.abstract_plotters import Plotter
from autogalaxy.plot.mat_plot.two_d import MatPlot2D
from autogalaxy.plot.visuals.two_d import Visuals2D
from autogalaxy.plot.include.two_d import Include2D


# TODO : Ew, this is a mass, but it works. Clean up one day!


class FitQuantityPlotter(Plotter):
    def __init__(
        self,
        fit: FitQuantity,
        mat_plot_2d: MatPlot2D = MatPlot2D(),
        visuals_2d: Visuals2D = Visuals2D(),
        include_2d: Include2D = Include2D(),
    ):
        """
        Plots the attributes of `FitQuantity` objects using the matplotlib method `imshow()` and many
        other matplotlib functions which customize the plot's appearance.

        The `mat_plot_1d` and `mat_plot_2d` attributes wrap matplotlib function calls to make the figure. By default,
        the settings passed to every matplotlib function called are those specified in
        the `config/visualize/mat_wrap/*.ini` files, but a user can manually input values into `MatPlot2d` to
        customize the figure's appearance.

        Overlaid on the figure are visuals, contained in the `Visuals1D` and `Visuals2D` objects. Attributes may be
        extracted from the `FitQuantity` and plotted via the visuals object, if the corresponding entry is `True` in
        the `Include1D` or `Include2D` object or the `config/visualize/include.ini` file.

        Parameters
        ----------
        fit
            The fit to an interferometer dataset the plotter plots.
        mat_plot_2d
            Contains objects which wrap the matplotlib function calls that make 2D plots.
        visuals_2d
            Contains 2D visuals that can be overlaid on 2D plots.
        include_2d
            Specifies which attributes of the `FitQuantity` are extracted and plotted as visuals for 2D plots.
        """
        super().__init__(
            mat_plot_2d=mat_plot_2d, include_2d=include_2d, visuals_2d=visuals_2d
        )

        self.fit = fit

    def get_visuals_2d(self) -> Visuals2D:
        return self.get_2d.via_fit_imaging_from(fit=self.fit)

    def get_visuals_2d_y(self) -> Visuals2D:
        return self.get_2d.via_fit_imaging_from(fit=self.fit.y)

    def get_visuals_2d_x(self) -> Visuals2D:
        return self.get_2d.via_fit_imaging_from(fit=self.fit.x)

    def figures_2d(
        self,
        image: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        model_image: bool = False,
        residual_map: bool = False,
        normalized_residual_map: bool = False,
        chi_squared_map: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `FitImaging` object in 2D.

        The API is such that every plottable attribute of the `FitImaging` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        image
            Whether to make a 2D plot (via `imshow`) of the image data.
        noise_map
            Whether to make a 2D plot (via `imshow`) of the noise map.
        signal_to_noise_map
            Whether to make a 2D plot (via `imshow`) of the signal-to-noise map.
        model_image
            Whether to make a 2D plot (via `imshow`) of the model image.
        residual_map
            Whether to make a 2D plot (via `imshow`) of the residual map.
        normalized_residual_map
            Whether to make a 2D plot (via `imshow`) of the normalized residual map.
        chi_squared_map
            Whether to make a 2D plot (via `imshow`) of the chi-squared map.
        """

        if isinstance(self.fit.dataset.data, aa.Array2D):
            fit_plotter = FitImagingPlotterMeta(
                fit=self.fit,
                get_visuals_2d=self.get_visuals_2d,
                mat_plot_2d=self.mat_plot_2d,
                include_2d=self.include_2d,
                visuals_2d=self.visuals_2d,
            )

            fit_plotter.figures_2d(
                data=image,
                noise_map=noise_map,
                signal_to_noise_map=signal_to_noise_map,
                model_image=model_image,
                residual_map=residual_map,
                normalized_residual_map=normalized_residual_map,
                chi_squared_map=chi_squared_map,
            )

        else:
            fit_plotter_y = FitImagingPlotterMeta(
                fit=self.fit.y,
                get_visuals_2d=self.get_visuals_2d_y,
                mat_plot_2d=self.mat_plot_2d,
                include_2d=self.include_2d,
                visuals_2d=self.visuals_2d,
            )

            fit_plotter_y.figures_2d(
                data=image,
                noise_map=noise_map,
                signal_to_noise_map=signal_to_noise_map,
                model_image=model_image,
                residual_map=residual_map,
                normalized_residual_map=normalized_residual_map,
                chi_squared_map=chi_squared_map,
                suffix="_y",
            )

            fit_plotter_x = FitImagingPlotterMeta(
                fit=self.fit.y,
                get_visuals_2d=self.get_visuals_2d_x,
                mat_plot_2d=self.mat_plot_2d,
                include_2d=self.include_2d,
                visuals_2d=self.visuals_2d,
            )

            fit_plotter_x.figures_2d(
                data=image,
                noise_map=noise_map,
                signal_to_noise_map=signal_to_noise_map,
                model_image=model_image,
                residual_map=residual_map,
                normalized_residual_map=normalized_residual_map,
                chi_squared_map=chi_squared_map,
                suffix="_x",
            )

    def subplot_fit(self):
        """
        Standard subplot of the attributes of the plotter's `FitQuantity` object.
        """

        if isinstance(self.fit.dataset.data, aa.Array2D):
            fit_plotter = FitImagingPlotterMeta(
                fit=self.fit,
                get_visuals_2d=self.get_visuals_2d,
                mat_plot_2d=self.mat_plot_2d,
                include_2d=self.include_2d,
                visuals_2d=self.visuals_2d,
            )

            fit_plotter.subplot(
                data=True,
                signal_to_noise_map=True,
                model_image=True,
                residual_map=True,
                normalized_residual_map=True,
                chi_squared_map=True,
                auto_filename="subplot_fit",
            )

        else:
            fit_plotter_y = FitImagingPlotterMeta(
                fit=self.fit.y,
                get_visuals_2d=self.get_visuals_2d_y,
                mat_plot_2d=self.mat_plot_2d,
                include_2d=self.include_2d,
                visuals_2d=self.visuals_2d,
            )

            fit_plotter_y.subplot(
                data=True,
                signal_to_noise_map=True,
                model_image=True,
                residual_map=True,
                normalized_residual_map=True,
                chi_squared_map=True,
                auto_filename="subplot_fit_y",
            )

            fit_plotter_x = FitImagingPlotterMeta(
                fit=self.fit.x,
                get_visuals_2d=self.get_visuals_2d_x,
                mat_plot_2d=self.mat_plot_2d,
                include_2d=self.include_2d,
                visuals_2d=self.visuals_2d,
            )

            fit_plotter_x.subplot(
                data=True,
                signal_to_noise_map=True,
                model_image=True,
                residual_map=True,
                normalized_residual_map=True,
                chi_squared_map=True,
                auto_filename="subplot_fit_x",
            )



from typing import Tuple

from autogalaxy.profiles.light.standard.sersic import Sersic


class DevVaucouleurs(Sersic):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
    ):
        """
        The elliptical Dev Vaucouleurs light profile.

        This is a specific case of the elliptical Sersic profile where `sersic_index=4.0`.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        """
        super().__init__(
            centre=centre,
            ell_comps=ell_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=4.0,
        )


class DevVaucouleursSph(DevVaucouleurs):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
    ):
        """
        The spherical Dev Vaucouleurs light profile.

        This is a specific case of the elliptical Sersic profile where `sersic_index=4.0`.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        effective_radius
            The circular radius containing half the light of this profile.
        """
        super().__init__(
            centre=centre,
            ell_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
        )



from typing import Tuple

from autogalaxy.profiles.light.snr.abstract import LightProfileSNR
from autogalaxy.profiles.light import standard as lp


class Gaussian(lp.Gaussian, LightProfileSNR):
    def __init__(
        self,
        signal_to_noise_ratio: float = 10.0,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        sigma: float = 1.0,
    ):
        """
        The elliptical Gaussian light profile.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        sigma
            The sigma value of the Gaussian, corresponding to ~ 1 / sqrt(2 log(2)) the full width half maximum.
        """

        super().__init__(centre=centre, ell_comps=ell_comps, intensity=0.0, sigma=sigma)
        LightProfileSNR.__init__(self, signal_to_noise_ratio=signal_to_noise_ratio)


class GaussianSph(lp.GaussianSph, LightProfileSNR):
    def __init__(
        self,
        signal_to_noise_ratio: float = 10.0,
        centre: Tuple[float, float] = (0.0, 0.0),
        sigma: float = 1.0,
    ):
        """
        The spherical Gaussian light profile.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        intensity
            Overall intensity normalisation of the light profile (units are dimensionless and derived from the data
            the light profile's image is compared too, which is expected to be electrons per second).
        sigma
            The sigma value of the Gaussian, corresponding to ~ 1 / sqrt(2 log(2)) the full width half maximum.
        """
        super().__init__(centre=centre, intensity=0.0, sigma=sigma)
        LightProfileSNR.__init__(self, signal_to_noise_ratio=signal_to_noise_ratio)



from typing import List

from autogalaxy.analysis.result import ResultDataset
from autogalaxy.ellipse.ellipse.ellipse import Ellipse
from autogalaxy.ellipse.fit_ellipse import FitEllipse


class ResultEllipse(ResultDataset):
    @property
    def max_log_likelihood_fit_list(self) -> FitEllipse:
        """
        An instance of a `FitEllipse` corresponding to the maximum log likelihood model inferred by the non-linear
        search.
        """
        return self.analysis.fit_list_from(instance=self.instance)

    @property
    def max_log_likelihood_ellipses(self) -> List[Ellipse]:
        """
        An instance of galaxies corresponding to the maximum log likelihood model inferred by the non-linear search.

        The galaxies list is computed from the `max_log_likelihood_fit`, as this ensures that all linear light profiles
        are converted to normal light profiles with their `intensity` values updated.
        """
        return self.instance.ellipses



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, List

if TYPE_CHECKING:
    from autogalaxy.interferometer.fit_interferometer import FitInterferometer

import autofit as af
import autoarray as aa

from autogalaxy.analysis.preloads import Preloads

from autogalaxy.aggregator import agg_util
from autogalaxy.aggregator.interferometer.interferometer import _interferometer_from
from autogalaxy.aggregator.dataset_model import _dataset_model_from
from autogalaxy.aggregator.galaxies import _galaxies_from


def _fit_interferometer_from(
    fit: af.Fit,
    instance: Optional[af.ModelInstance] = None,
    real_space_mask: Optional[aa.Mask2D] = None,
    settings_inversion: aa.SettingsInversion = None,
    use_preloaded_grid: bool = True,
) -> List[FitInterferometer]:
    """
    Returns a list of `FitInterferometer` objects from a `PyAutoFit` sqlite database `Fit` object.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The interferometer data, noise-map, uv-wavelengths and settings as .fits files (e.g. `dataset/data.fits`).
    - The real space mask defining the grid of the interferometer for the FFT (`dataset/real_space_mask.fits`).
    - The settings of inversions used by the fit (`dataset/settings_inversion.json`).

    Each individual attribute can be loaded from the database via the `fit.value()` method.

    This method combines all of these attributes and returns a `FitInterferometer` object for a given non-linear
    search sample (e.g. the maximum likelihood model). This includes associating adapt images with their respective
    galaxies.

    If multiple `FitInterferometer` objects were fitted simultaneously via analysis summing, the `fit.child_values()`
    method is instead used to load lists of the data, noise-map, PSF and mask and combine them into a list of
    `FitInterferometer` objects.

    The settings of an inversion can be overwritten by inputting a `settings_inversion` object, for
    example if you want to use a grid with a different inversion solver.

    Parameters
    ----------
    fit
        A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
    instance
        A manual instance that overwrites the max log likelihood instance in fit (e.g. for drawing the instance
        randomly from the PDF).
    settings_inversion
        Optionally overwrite the `SettingsInversion` of the `Inversion` object that is created from the fit.
    use_preloaded_grid
        Certain pixelization's construct their mesh in the source-plane from a stochastic KMeans algorithm. This grid
        may be output to hard-disk after the model-fit and loaded via the database to ensure the same grid is used
        as the fit.
    """
    from autogalaxy.interferometer.fit_interferometer import FitInterferometer

    dataset_list = _interferometer_from(
        fit=fit,
        real_space_mask=real_space_mask,
    )

    galaxies_list = _galaxies_from(fit=fit, instance=instance)

    dataset_model_list = _dataset_model_from(fit=fit, instance=instance)

    adapt_images_list = agg_util.adapt_images_from(fit=fit)

    settings_inversion = settings_inversion or fit.value(name="settings_inversion")

    mesh_grids_of_planes_list = agg_util.mesh_grids_of_planes_list_from(
        fit=fit, total_fits=len(dataset_list), use_preloaded_grid=use_preloaded_grid
    )

    fit_dataset_list = []

    for dataset, galaxies, dataset_model, adapt_images, mesh_grids_of_planes in zip(
        dataset_list,
        galaxies_list,
        dataset_model_list,
        adapt_images_list,
        mesh_grids_of_planes_list,
    ):
        preloads = agg_util.preloads_from(
            preloads_cls=Preloads,
            use_preloaded_grid=use_preloaded_grid,
            mesh_grids_of_planes=mesh_grids_of_planes,
            use_w_tilde=False,
        )

        fit_dataset_list.append(
            FitInterferometer(
                dataset=dataset,
                galaxies=galaxies,
                dataset_model=dataset_model,
                adapt_images=adapt_images,
                settings_inversion=settings_inversion,
                preloads=preloads,
            )
        )

    return fit_dataset_list


class FitInterferometerAgg(af.AggBase):
    def __init__(
        self,
        aggregator: af.Aggregator,
        settings_inversion: Optional[aa.SettingsInversion] = None,
        use_preloaded_grid: bool = True,
        real_space_mask: Optional[aa.Mask2D] = None,
    ):
        """
        Interfaces with an `PyAutoFit` aggregator object to create instances of `FitInterferometer` objects from the
        results of a model-fit.

        The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

        - The interferometer data, noise-map, uv-wavelengths and settings as .fits files (e.g. `dataset/data.fits`).
        - The real space mask defining the grid of the interferometer for the FFT (`dataset/real_space_mask.fits`).
        - The settings of inversions used by the fit (`dataset/settings_inversion.json`).

        The `aggregator` contains the path to each of these files, and they can be loaded individually. This class
        can load them all at once and create an `FitInterferometer` object via the `_fit_interferometer_from` method.

        This class's methods returns generators which create the instances of the `FitInterferometer` objects. This ensures
        that large sets of results can be efficiently loaded from the hard-disk and do not require storing all
        `FitInterferometer` instances in the memory at once.

        For example, if the `aggregator` contains 3 model-fits, this class can be used to create a generator which
        creates instances of the corresponding 3 `FitInterferometer` objects.

        If multiple `Imaging` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
        is instead used to load lists of the data, noise-map, PSF and mask and combine them into a list of
        `FitImaging` objects.

        This can be done manually, but this object provides a more concise API.

        Parameters
        ----------
        aggregator
            A `PyAutoFit` aggregator object which can load the results of model-fits.
        settings_inversion
            Optionally overwrite the `SettingsInversion` of the `Inversion` object that is created from the fit.
        use_preloaded_grid
            Certain pixelization's construct their mesh in the source-plane from a stochastic KMeans algorithm. This
            grid may be output to hard-disk after the model-fit and loaded via the database to ensure the same grid is
            used as the fit.
        """
        super().__init__(aggregator=aggregator)

        self.settings_inversion = settings_inversion
        self.use_preloaded_grid = use_preloaded_grid
        self.real_space_mask = real_space_mask

    def object_via_gen_from(
        self, fit, instance: Optional[af.ModelInstance] = None
    ) -> List[FitInterferometer]:
        """
        Returns a generator of `FitInterferometer` objects from an input aggregator.

        See `__init__` for a description of how the `FitInterferometer` objects are created by this method.

        Parameters
        ----------
        fit
            A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
        instance
            A manual instance that overwrites the max log likelihood instance in fit (e.g. for drawing the instance
            randomly from the PDF).
        """
        return _fit_interferometer_from(
            fit=fit,
            instance=instance,
            settings_inversion=self.settings_inversion,
            use_preloaded_grid=self.use_preloaded_grid,
        )



from typing import List, Optional

import autoarray as aa
import autoarray.plot as aplt

from autogalaxy.plot.abstract_plotters import Plotter
from autogalaxy.plot.mat_plot.one_d import MatPlot1D
from autogalaxy.plot.mat_plot.two_d import MatPlot2D
from autogalaxy.plot.visuals.one_d import Visuals1D
from autogalaxy.plot.visuals.two_d import Visuals2D
from autogalaxy.plot.include.one_d import Include1D
from autogalaxy.plot.include.two_d import Include2D
from autogalaxy.plot.mass_plotter import MassPlotter
from autogalaxy.galaxy.galaxy import Galaxy
from autogalaxy.galaxy.galaxies import Galaxies
from autogalaxy.galaxy.plot.galaxy_plotters import GalaxyPlotter

from autogalaxy import exc


class GalaxiesPlotter(Plotter):
    def __init__(
        self,
        galaxies: List[Galaxy],
        grid: aa.type.Grid1D2DLike,
        mat_plot_1d: MatPlot1D = MatPlot1D(),
        visuals_1d: Visuals1D = Visuals1D(),
        include_1d: Include1D = Include1D(),
        mat_plot_2d: MatPlot2D = MatPlot2D(),
        visuals_2d: Visuals2D = Visuals2D(),
        include_2d: Include2D = Include2D(),
    ):
        """
        Plots the attributes of a list of galaxies using the matplotlib methods `plot()` and `imshow()` and many
        other matplotlib functions which customize the plot's appearance.

        The `mat_plot_1d` and `mat_plot_2d` attributes wrap matplotlib function calls to make the figure. By default,
        the settings passed to every matplotlib function called are those specified in
        the `config/visualize/mat_wrap/*.ini` files, but a user can manually input values into `MatPlot2D` to
        customize the figure's appearance.

        Overlaid on the figure are visuals, contained in the `Visuals1D` and `Visuals2D` objects. Attributes may be
        extracted from the `MassProfile` and plotted via the visuals object, if the corresponding entry is `True` in
        the `Include1D` or `Include2D` object or the `config/visualize/include.ini` file.

        Parameters
        ----------
        galaxies
            The galaxies the plotter plots.
        grid
            The 2D (y,x) grid of coordinates used to evaluate the galaxies light and mass quantities that are plotted.
        mat_plot_1d
            Contains objects which wrap the matplotlib function calls that make 1D plots.
        visuals_1d
            Contains 1D visuals that can be overlaid on 1D plots.
        include_1d
            Specifies which attributes of the `MassProfile` are extracted and plotted as visuals for 1D plots.
        mat_plot_2d
            Contains objects which wrap the matplotlib function calls that make 2D plots.
        visuals_2d
            Contains 2D visuals that can be overlaid on 2D plots.
        include_2d
            Specifies which attributes of the `MassProfile` are extracted and plotted as visuals for 2D plots.
        """

        self.galaxies = Galaxies(galaxies=galaxies)

        from autogalaxy.profiles.light.linear import (
            LightProfileLinear,
        )

        if self.galaxies.has(cls=LightProfileLinear):
            raise exc.raise_linear_light_profile_in_plot(
                plotter_type=self.__class__.__name__,
            )

        super().__init__(
            mat_plot_2d=mat_plot_2d,
            include_2d=include_2d,
            visuals_2d=visuals_2d,
            mat_plot_1d=mat_plot_1d,
            include_1d=include_1d,
            visuals_1d=visuals_1d,
        )

        self.grid = grid

        self._mass_plotter = MassPlotter(
            mass_obj=self.galaxies,
            grid=self.grid,
            get_visuals_2d=self.get_visuals_2d,
            mat_plot_2d=self.mat_plot_2d,
            include_2d=self.include_2d,
            visuals_2d=self.visuals_2d,
        )

    def get_visuals_2d(self) -> Visuals2D:
        return self.get_2d.via_light_mass_obj_from(
            light_mass_obj=self.galaxies, grid=self.grid
        )

    def get_visuals_2d_of_galaxy(self, galaxy_index: int) -> aplt.Visuals2D:
        return self.get_2d.via_galaxies_from(
            galaxies=self.galaxies, grid=self.grid, galaxy_index=galaxy_index
        )

    def galaxy_plotter_from(self, galaxy_index: int) -> GalaxyPlotter:
        """
        Returns an `GalaxyPlotter` corresponding to a `Galaxy` in the `Tracer`.

        Returns
        -------
        galaxy_index
            The index of the galaxy in the `Tracer` used to make the `GalaxyPlotter`.
        """

        return GalaxyPlotter(
            galaxy=self.galaxies[galaxy_index],
            grid=self.grid,
            mat_plot_2d=self.mat_plot_2d,
            visuals_2d=self.get_visuals_2d_of_galaxy(galaxy_index=galaxy_index),
            include_2d=self.include_2d,
        )

    def figures_2d(
        self,
        image: bool = False,
        convergence: bool = False,
        potential: bool = False,
        deflections_y: bool = False,
        deflections_x: bool = False,
        magnification: bool = False,
        plane_image: bool = False,
        plane_grid: bool = False,
        zoom_to_brightest: bool = True,
        title_suffix: str = "",
        filename_suffix: str = "",
        source_plane_title: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `Galaxies` object in 2D, which are computed via the plotter's 2D
        grid object.

        The API is such that every plottable attribute of the `Galaxies` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        image
            Whether to make a 2D plot (via `imshow`) of the image of the galaxies.
        convergence
            Whether to make a 2D plot (via `imshow`) of the convergence.
        potential
            Whether to make a 2D plot (via `imshow`) of the potential.
        deflections_y
            Whether to make a 2D plot (via `imshow`) of the y component of the deflection angles.
        deflections_x
            Whether to make a 2D plot (via `imshow`) of the x component of the deflection angles.
        magnification
            Whether to make a 2D plot (via `imshow`) of the magnification.
        plane_image
            Whether to make a 2D plot (via `imshow`) of the image of the plane in the soure-plane (e.g. its
            unlensed light).
        zoom_to_brightest
            Whether to automatically zoom the plot to the brightest regions of the galaxies being plotted as
            opposed to the full extent of the grid.
        title_suffix
            Add a suffix to the end of the matplotlib title label.
        filename_suffix
            Add a suffix to the end of the filename the plot is saved to hard-disk using.
        """
        if image:
            self.mat_plot_2d.plot_array(
                array=self.galaxies.image_2d_from(grid=self.grid),
                visuals_2d=self.get_visuals_2d(),
                auto_labels=aplt.AutoLabels(
                    title=f"Image{title_suffix}", filename=f"image_2d{filename_suffix}"
                ),
            )

        if plane_image:
            if source_plane_title:
                title = "Source Plane Image"
            else:
                title = f"Plane Image{title_suffix}"

            self.mat_plot_2d.plot_array(
                array=self.galaxies.plane_image_2d_from(
                    grid=self.grid, zoom_to_brightest=zoom_to_brightest
                ),
                visuals_2d=self.get_visuals_2d(),
                auto_labels=aplt.AutoLabels(
                    title=title,
                    filename=f"plane_image{filename_suffix}",
                ),
            )

        if plane_grid:
            if source_plane_title:
                title = "Source Plane Grid"
            else:
                title = f"Plane Grid{title_suffix}"

            self.mat_plot_2d.plot_grid(
                grid=self.grid,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=aplt.AutoLabels(
                    title=title,
                    filename=f"plane_grid{filename_suffix}",
                ),
            )

        self._mass_plotter.figures_2d(
            convergence=convergence,
            potential=potential,
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            magnification=magnification,
        )

    def galaxy_indexes_from(self, galaxy_index: Optional[int]) -> List[int]:
        """
        Returns a list of all indexes of the galaxys in the fit, which is iterated over in figures that plot
        individual figures of each galaxy.

        Parameters
        ----------
        galaxy_index
            A specific galaxy index which when input means that only a single galaxy index is returned.

        Returns
        -------
        list
            A list of galaxy indexes corresponding to galaxys in the galaxy.
        """
        if galaxy_index is None:
            return list(range(len(self.galaxies)))
        return [galaxy_index]

    def figures_2d_of_galaxies(
        self, image: bool = False, galaxy_index: Optional[int] = None
    ):
        """
        Plots galaxy images for each individual `Galaxy` in the plotter's `Galaxies` in 2D,  which are computed via the
        plotter's 2D grid object.

        The API is such that every plottable attribute of the `galaxy` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        image
            Whether to make a 2D plot (via `imshow`) of the image of the galaxy in the soure-galaxy (e.g. its
            unlensed light).
        galaxy_index
            If input, plots for only a single galaxy based on its index are created.
        """
        galaxy_indexes = self.galaxy_indexes_from(galaxy_index=galaxy_index)

        for galaxy_index in galaxy_indexes:
            galaxy_plotter = self.galaxy_plotter_from(galaxy_index=galaxy_index)

            if image:
                galaxy_plotter.figures_2d(
                    image=True,
                    title_suffix=f" Of Galaxy {galaxy_index}",
                    filename_suffix=f"_of_galaxy_{galaxy_index}",
                )

    def subplot(
        self,
        image: bool = False,
        convergence: bool = False,
        potential: bool = False,
        deflections_y: bool = False,
        deflections_x: bool = False,
        magnification: bool = False,
        auto_filename: str = "subplot_galaxies",
    ):
        """
        Plots the individual attributes of the plotter's `Galaxies` object in 2D on a subplot, which are computed via the
        plotter's 2D grid object.

        The API is such that every plottable attribute of the `Galaxies` object is an input parameter of type bool of
        the function, which if switched to `True` means that it is included on the subplot.

        Parameters
        ----------
        image
            Whether or not to include a 2D plot (via `imshow`) of the image.
        convergence
            Whether or not to include a 2D plot (via `imshow`) of the convergence.
        potential
            Whether or not to include a 2D plot (via `imshow`) of the potential.
        deflections_y
            Whether or not to include a 2D plot (via `imshow`) of the y component of the deflection angles.
        deflections_x
            Whether or not to include a 2D plot (via `imshow`) of the x component of the deflection angles.
        magnification
            Whether or not to include a 2D plot (via `imshow`) of the magnification.
        auto_filename
            The default filename of the output subplot if written to hard-disk.
        """
        self._subplot_custom_plot(
            image=image,
            convergence=convergence,
            potential=potential,
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            magnification=magnification,
            auto_labels=aplt.AutoLabels(filename=auto_filename),
        )

    def subplot_galaxies(self):
        """
        Standard subplot of the attributes of the plotter's `Galaxies` object.
        """
        return self.subplot(
            image=True,
            convergence=True,
            potential=True,
            deflections_y=True,
            deflections_x=True,
        )

    def subplot_galaxy_images(self):
        """
        Subplot of the image of every galaxy.

        For example, for a 2 galaxy `Galaxies`, this creates a subplot with 2 panels, one for each galaxy.
        """
        number_subplots = len(self.galaxies)

        self.open_subplot_figure(number_subplots=number_subplots)

        for galaxy_index in range(0, len(self.galaxies)):
            galaxy_plotter = self.galaxy_plotter_from(galaxy_index=galaxy_index)
            galaxy_plotter.figures_2d(
                image=True, title_suffix=f" Of Galaxies {galaxy_index}"
            )

        self.mat_plot_2d.output.subplot_to_figure(
            auto_filename=f"subplot_galaxy_images"
        )
        self.close_subplot_figure()



from autogalaxy import exc
from autogalaxy.profiles.mass.dark.abstract import DarkProfile
from autogalaxy.profiles.mass.stellar.abstract import StellarProfile


class StellarDarkDecomp:
    def __init__(self, galaxy):
        self.galaxy = galaxy

    def stellar_mass_angular_within_circle_from(self, radius: float):
        if self.galaxy.has(cls=StellarProfile):
            return sum(
                [
                    profile.mass_angular_within_circle_from(radius=radius)
                    for profile in self.galaxy.cls_list_from(cls=StellarProfile)
                ]
            )
        else:
            raise exc.GalaxyException(
                "You cannot perform a stellar mass-based calculation on a galaxy which does not have a stellar "
                "mass-profile "
            )

    def dark_mass_angular_within_circle_from(self, radius: float):
        if self.galaxy.has(cls=DarkProfile):
            return sum(
                [
                    profile.mass_angular_within_circle_from(radius=radius)
                    for profile in self.galaxy.cls_list_from(cls=DarkProfile)
                ]
            )
        else:
            raise exc.GalaxyException(
                "You cannot perform a dark mass-based calculation on a galaxy which does not have a dark mass-profile"
            )

    def stellar_fraction_at_radius_from(self, radius):
        return 1.0 - self.dark_fraction_at_radius_from(radius=radius)

    def dark_fraction_at_radius_from(self, radius):
        stellar_mass = self.stellar_mass_angular_within_circle_from(radius=radius)
        dark_mass = self.dark_mass_angular_within_circle_from(radius=radius)

        return dark_mass / (stellar_mass + dark_mass)



from typing import Tuple

from autogalaxy.profiles.light.linear.abstract import LightProfileLinear

from autogalaxy.profiles.light import standard as lp


class Gaussian(lp.Gaussian, LightProfileLinear):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        ell_comps: Tuple[float, float] = (0.0, 0.0),
        sigma: float = 1.0,
    ):
        """
        The elliptical Gaussian light profile.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        ell_comps
            The first and second ellipticity components of the elliptical coordinate system.
        sigma
            The sigma value of the Gaussian, corresponding to ~ 1 / sqrt(2 log(2)) the full width half maximum.
        """

        super().__init__(centre=centre, ell_comps=ell_comps, intensity=1.0, sigma=sigma)


class GaussianSph(Gaussian):
    def __init__(
        self,
        centre: Tuple[float, float] = (0.0, 0.0),
        sigma: float = 1.0,
    ):
        """
        The spherical Gaussian light profile.

        Parameters
        ----------
        centre
            The (y,x) arc-second coordinates of the profile centre.
        sigma
            The sigma value of the Gaussian, corresponding to ~ 1 / sqrt(2 log(2)) the full width half maximum.
        """
        super().__init__(centre=centre, ell_comps=(0.0, 0.0), sigma=sigma)
